# copyright 2016-2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""cubicweb-jsonschema entity's classes"""

from logilab.common.registry import objectify_predicate

from cubicweb.predicates import match_kwargs, non_final_entity
from cubicweb.entity import Adapter, EntityAdapter

from cubicweb_jsonschema import (
    VIEW_ROLE,
    CREATION_ROLE,
    EDITION_ROLE,
    orm_rtype,
)


def jsonschema_adapter(cnx, **ctx):
    """Return a IJSONSchema adapter selected from ``ctx`` information."""
    return cnx.vreg["adapters"].select("IJSONSchema", cnx, **ctx)


class PostCreateMixin:
    """Handles reverse dependant entities post creation."""

    def _pop_post_creates(self, values):
        post_creates = []
        for key in list(values.keys()):
            if key.startswith("__post_create"):
                post_create_elt = values.pop(key)
                if isinstance(post_create_elt, list):
                    post_creates.extend(post_create_elt)
                else:
                    post_creates.append(post_create_elt)
        return post_creates

    def _handle_post_creates(self, post_creates, entity_eid):
        for post_create in post_creates:
            relation = post_create["relation"]
            post_create["values"][relation] = entity_eid
            self._cw.create_entity(post_create["type"], **post_create["values"])


class IJSONSchemaETypeAdapter(Adapter, PostCreateMixin):
    """IJSONSchema adapter for entity creation with `etype` specified in
    selection context.

    Subclasses may refine the selection context using ``match_kwargs``
    predicates (e.g. ``match_kwargs({'etype': 'MyEtype'})`` to override JSON
    Schema generation for a particular etype).
    """

    __regid__ = "IJSONSchema"
    __select__ = match_kwargs("etype")

    def __repr__(self):
        return "<{0.__class__.__name__} etype={0.etype}>".format(self)

    @property
    def etype(self):
        """The entity type bound to this adapter."""
        return str(self.cw_extra_kwargs["etype"])

    @property
    def entity_mapper(self):
        return self._cw.vreg["mappers"].select(
            "jsonschema.entity", self._cw, etype=self.etype
        )

    def view_schema(self, **kwargs):
        """Return a JSON schema suitable for viewing adapted entity."""
        jsldoc = self.entity_mapper.jsl_document(VIEW_ROLE)
        return jsldoc.get_schema(**kwargs)

    def creation_schema(self, **kwargs):
        """Return a JSON schema suitable for entity creation."""
        jsldoc = self.entity_mapper.jsl_document(CREATION_ROLE)
        return jsldoc.get_schema(**kwargs)

    @property
    def entity(self):
        return self._cw.vreg["etypes"].etype_class(self.etype)(self._cw)

    def create_entity(self, instance):
        """Return a CubicWeb entity built from `instance` data matching this
        JSON schema.
        """
        values = self.entity_mapper.values(instance)
        post_creates = self._pop_post_creates(values)
        entity = self._cw.create_entity(self.etype, **values)
        self._handle_post_creates(post_creates, entity.eid)
        return entity

    def serialize(self, entities):
        """Return `entities` as a collection for JSON serialization."""
        mapper = self._cw.vreg["mappers"].select(
            "jsonschema.collection", self._cw, etype=self.etype
        )
        return mapper.serialize(entities)


class IJSONSchemaRelationTargetETypeAdapter(IJSONSchemaETypeAdapter):
    """IJSONSchema adapter for entity creation of an entity of `etype` related
    through `rtype` and `role` specified in selection context.
    """

    __select__ = match_kwargs("etype", "rtype", "role")

    def __repr__(self):
        return (
            "<{0.__class__.__name__} etype={0.etype} rtype={0.rtype} "
            "role={0.role}>".format(self)
        )

    @property
    def rtype(self):
        return self.cw_extra_kwargs["rtype"]

    @property
    def role(self):
        return self.cw_extra_kwargs["role"]

    @property
    def entity_mapper(self):
        return self._cw.vreg["mappers"].select(
            "jsonschema.entity",
            self._cw,
            etype=self.etype,
            rtype=self.rtype,
            role=self.role,
        )

    def create_entity(self, instance, target=None):
        """Return a CubicWeb entity related to `target` through relation
        information from selection context.
        """
        entity = super(
            IJSONSchemaRelationTargetETypeAdapter, self
        ).create_entity(instance)
        if target is not None:
            entity.cw_set(**{orm_rtype(self.rtype, self.role): target})
        return entity


class IJSONSchemaEntityAdapter(EntityAdapter, PostCreateMixin):
    """IJSONSchema adapter for live entities."""

    __regid__ = "IJSONSchema"
    # Prevent this adapter from being selected in place of
    # IJSONSchemaRelatedEntityAdapter in case one picks a bad rtype/role.
    __select__ = non_final_entity() & ~match_kwargs("rtype", "role")

    def __repr__(self):
        return "<{0.__class__.__name__} entity={0.entity}>".format(self)

    @property
    def etype(self):
        return self.entity.cw_etype

    @property
    def entity_mapper(self):
        return self._cw.vreg["mappers"].select(
            "jsonschema.entity", self._cw, entity=self.entity
        )

    def view_schema(self, **kwargs):
        """Return a JSON schema suitable for viewing adapted entity."""
        jsldoc = self.entity_mapper.jsl_document(VIEW_ROLE)
        return jsldoc.get_schema(**kwargs)

    def edition_schema(self, **kwargs):
        """Return a JSON schema suitable for editing adapted entity."""
        jsldoc = self.entity_mapper.jsl_document(EDITION_ROLE)
        return jsldoc.get_schema(**kwargs)

    def relations(self):
        """Yield (rtype, role, targettypes) for non-final relations."""
        rsection = self._cw.vreg["uicfg"].select(
            "jsonschema", self._cw, entity=self.entity
        )
        return rsection.relations_by_section(self.entity, "related", "read")

    def edit_entity(self, instance):
        """Return a CubicWeb entity built from `instance` data matching this
        JSON schema.
        """
        values = self.entity_mapper.values(instance)
        post_creates = self._pop_post_creates(values)
        self.entity.cw_set(**values)
        self._handle_post_creates(post_creates, self.entity.eid)
        return self.entity

    def serialize(self):
        """Return a dictionary of entity's data suitable for JSON
        serialization.
        """
        return self.entity_mapper.serialize()


@objectify_predicate
def relation_for_entity(cls, cnx, entity, rtype, role):
    """Return 1 if `entity` has an `(rtype, role)` relation."""
    return 1 if entity.e_schema.has_relation(rtype, role) else 0


class IJSONSchemaRelatedEntityAdapter(IJSONSchemaEntityAdapter):
    """IJSONSchema adapter for entities in the context of a relation."""

    __select__ = (
        non_final_entity()
        & match_kwargs("rtype", "role")
        & relation_for_entity()
    )

    def __repr__(self):
        return (
            "<{0.__class__.__name__} entity={0.entity} "
            "rtype={0.rtype} role={0.role}>".format(self)
        )

    @property
    def rtype(self):
        return self.cw_extra_kwargs["rtype"]

    @property
    def role(self):
        return self.cw_extra_kwargs["role"]



# copyright 2016-2022 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""Appobjects for mapping Yams schema definitions to JSON Schema documents."""

from collections import OrderedDict
from functools import partial

import iso8601

import jsl

from logilab.common.decorators import cachedproperty, classproperty
from logilab.common.registry import Predicate, objectify_predicate, yes

from yams import BadSchemaDefinition, ValidationError
from yams.constraints import StaticVocabularyConstraint
from cubicweb import Binary, neg_role, _
from cubicweb.appobject import AppObject
from cubicweb.predicates import (
    ExpectedValuePredicate,
    PartialPredicateMixIn,
    match_kwargs,
)

from cubicweb_jsonschema import (
    CREATION_ROLE,
    EDITION_ROLE,
    VIEW_ROLE,
    orm_rtype,
)
from cubicweb_jsonschema.views import jsonschema_section


def _etype_from_context(kwargs):
    etype = kwargs.get("etype")
    if etype is None:
        entity = kwargs.get("entity")
        if entity is not None:
            etype = entity.cw_etype
    return etype


@objectify_predicate
def yams_final_rtype(cls, cnx, rtype, role, **kwargs):
    """Predicate that returns 1 when the supplied `rtype` is not final."""
    etype = _etype_from_context(kwargs)
    if etype is None:
        return 0
    rdef = cnx.vreg.schema[etype].relation_definition(
        rtype, role=role, takefirst=True
    )
    if rdef.final:
        return 1
    return 0


@objectify_predicate
def yams_component_target(cls, cnx, rtype, role, target_types=None, **kwargs):
    """Predicate that returns 1 when the target entity types are component of
    the relation defined by `rtype` and `role` (i.e. the relation has
    composite set to `role`).
    """
    etype = _etype_from_context(kwargs)
    if etype is None:
        return 0
    component = None
    eschema = cnx.vreg.schema[etype]
    if target_types is None:
        rschema = cnx.vreg.schema[rtype]
        target_types = tuple(t.type for t in rschema.targets(etype, role=role))
    for target_type in target_types:
        rdef = eschema.relation_definition(
            rtype, role=role, targettype=target_type
        )
        _component = rdef.composite == role
        if component is None:
            component = _component
        elif not component == _component:
            cls.warning("component inconsistency accross target types")
            return 0
    return component


class yams_match(Predicate):
    """Predicate that returns a positive value when the supplied relation
    match parameters given as predicate argument.

    The more __call__ arguments match __init__'s ones, the higher the score
    is.
    """

    def __init__(self, etype=None, rtype=None, role=None, target_types=()):
        self.etype = etype
        self.rtype = rtype
        self.role = role
        if isinstance(target_types, str):
            target_types = {target_types}
        elif not isinstance(target_types, (set, frozenset)):
            target_types = frozenset(target_types)
        self.target_types = target_types

    def __call__(self, cls, cnx, rtype, role, **kwargs):
        etype = _etype_from_context(kwargs)
        score = 0
        for key in ("etype", "rtype", "role"):
            expected = getattr(self, key, None)
            if expected is not None:
                if locals()[key] != expected:
                    return 0
                score += 1
        target_types = kwargs.get("target_types", frozenset())
        assert isinstance(
            target_types, (set, frozenset)
        ), 'yams_match must be called with a set as "target_types" argument'
        if self.target_types:
            if not target_types.issubset(self.target_types):
                return 0
            score += 1
        return score


class partial_yams_match(PartialPredicateMixIn, ExpectedValuePredicate):
    """Return non-zero if selected class's attributes `etype`, `rtype`, `role`
    and/or `target_types` match with keyword arguments from context.
    """

    match_attributes = ("etype", "rtype", "role", "target_types")

    def __init__(self, *expected, **kwargs):
        super(partial_yams_match, self).__init__((), **kwargs)

    def complete(self, cls):
        self.expected = {}
        for name in self.match_attributes:
            if hasattr(cls, name):
                self.expected[name] = set([getattr(cls, name)])

    def _values_set(self, cls, req, **kwargs):
        values = {}
        for name in self.match_attributes:
            if name in kwargs:
                values[name] = kwargs[name]
        return values


def base_fields(etype):
    fields = OrderedDict()
    fields["__etype__"] = jsl.fields.StringField(
        required=True, enum=[etype], default=etype
    )
    fields["__eid__"] = jsl.fields.StringField(required=True)
    return fields


def default_options(_cw, etype):
    """Return an Options class for a jsl.Document named `etype`."""
    attrs = {
        "definition_id": etype,
        "title": _cw._(etype),
        "schema_uri": "http://json-schema.org/draft-06/schema#",
    }
    return type("Options", (object,), attrs)


class JSONSchemaMapper(AppObject):
    """Base class for mappers between Yams schema definitions and JSON Schema
    documents.
    """

    __registry__ = "mappers"


class ETypeMapper(JSONSchemaMapper):
    """A mapper to build JSL document for an entity type."""

    __regid__ = "jsonschema.entity"
    __select__ = match_kwargs("etype")

    # Should "base" fields __etype__/__eid__ be included in main schema
    # document?
    include_base_fields = False

    @property
    def etype(self):
        return self.cw_extra_kwargs["etype"]

    @property
    def entity(self):
        """Entity bound to this mapper, None for ETypeMapper."""
        return None

    def jsl_document(self, schema_role=None):
        """Return a jsl.Document for bound entity type with fields relations
        registered to appear in the document (through `jsonschema_section`).
        """
        attrs = OrderedDict()
        if self.include_base_fields:
            attrs.update(base_fields(self.etype))
        for mapper in self._object_mappers():
            attrs[mapper.name] = mapper.jsl_field(schema_role)
        for rtype, role, target_types in self.relations(schema_role):
            if target_types is None:
                target_types = self._rtype_target_types(rtype, role)
            assert isinstance(target_types, set), target_types
            mapper = self._relation_mapper(rtype, role, target_types)
            attrs[rtype] = mapper.jsl_field(schema_role)
        attrs["Options"] = default_options(self._cw, self.etype)
        return type(str(self.etype), (jsl.Document,), attrs)

    def values(self, instance):
        """Return a dict with deserialized data from `instance` suitable for
        insertion in CubicWeb database.
        """
        entity = self.entity
        values = {}
        # Deserialize "jsonschema.object" mappers first.
        for mapper in self._object_mappers():
            values.update(mapper.values(entity, instance))
        # Then Yams relations.
        if entity is None:
            schema_role = CREATION_ROLE
        else:
            schema_role = EDITION_ROLE
        for rtype, role, target_types in self.relations(schema_role):
            mapper = self._relation_mapper(rtype, role, target_types)
            values.update(mapper.values(entity, instance))
        if instance:
            # All properties in "instance" should have been consumed at this
            # point.
            eid = entity.eid if entity is not None else None
            msg = _("unexpected properties: {}").format(", ".join(instance))
            raise ValidationError(eid, {None: msg})
        return values

    def relations(self, schema_role):
        """Yield relation information tuple (rtype, role, targettypes)
        for given schema role in the context of bound entity type.
        """
        try:
            permission = {
                VIEW_ROLE: "read",
                CREATION_ROLE: "add",
                EDITION_ROLE: "update",
            }[schema_role]
        except KeyError:
            raise ValueError(
                'unhandled schema role "{0}" in {1}'.format(schema_role, self)
            )
        entity = self.entity
        if entity is None:
            entity = self._cw.vreg["etypes"].etype_class(self.etype)(self._cw)
        rsection = self._cw.vreg["uicfg"].select(
            "jsonschema", self._cw, entity=entity
        )
        return rsection.relations_by_section(entity, "inlined", permission)

    def _object_mappers(self):
        """Yield 'jsonschema.object' mapper instance selectable for entity
        bound to this mapper.
        """
        if "jsonschema.object" not in self._cw.vreg["mappers"]:
            return
        for mappercls in self._cw.vreg["mappers"]["jsonschema.object"]:
            if mappercls.__select__(mappercls, self._cw, etype=self.etype) > 0:
                yield mappercls(self._cw)

    def _rtype_target_types(self, rtype, role):
        rschema = self._cw.vreg.schema[rtype]
        return {t.type for t in rschema.targets(self.etype, role)}

    def _relation_mapper(self, rtype, role, target_types=None):
        return self._cw.vreg["mappers"].select(
            "jsonschema.relation",
            self._cw,
            etype=self.etype,
            rtype=rtype,
            role=role,
            target_types=target_types,
        )


class TargetETypeMapper(ETypeMapper):
    """Specialized version of :class:`ETypeMapper` selectable for an entity
    type as target of (`rtype`, `role`) relation.
    """

    __select__ = match_kwargs("etype", "rtype", "role")

    @property
    def rtype(self):
        return self.cw_extra_kwargs["rtype"]

    @property
    def role(self):
        return self.cw_extra_kwargs["role"]

    def relations(self, schema_role):
        relations = super(TargetETypeMapper, self).relations(schema_role)
        for rtype, role, target_types in relations:
            if (rtype, role) == (self.rtype, self.role):
                continue
            yield rtype, role, target_types


class EntityMapper(ETypeMapper):
    """Specialized version of :class:`ETypeMapper` for a live entity."""

    __select__ = match_kwargs("entity")

    @property
    def entity(self):
        """Live entity from selection context."""
        return self.cw_extra_kwargs["entity"]

    @property
    def etype(self):
        return self.entity.cw_etype

    def _relation_mapper(self, rtype, role, target_types=None):
        return self._cw.vreg["mappers"].select(
            "jsonschema.relation",
            self._cw,
            entity=self.entity,
            rtype=rtype,
            role=role,
            target_types=target_types,
        )

    def serialize(self):
        """Return the serialized value entity bound to this mapper."""
        entity = self.entity
        entity.complete()
        data = {}
        for mapper in self._object_mappers():
            data[mapper.name] = mapper.serialize(entity)
        for rtype, role, target_types in self.relations(VIEW_ROLE):
            relation_mapper = self._relation_mapper(rtype, role, target_types)
            value = relation_mapper.serialize(entity)
            if value is None:
                continue
            data[relation_mapper.orm_rtype] = value
        return data


class TargetEntityMapper(EntityMapper, TargetETypeMapper):
    """Specialized version of :class:`TargetETypeMapper` and
    :class:`EntityMapper` selectable for an *entity* as target of (`rtype`,
    `role`) relation.
    """

    __select__ = match_kwargs("entity", "rtype", "role")


class EntityCollectionMapper(JSONSchemaMapper):
    """Mapper for a collection of entities."""

    __regid__ = "jsonschema.collection"
    __select__ = match_kwargs("etype")

    @property
    def etype(self):
        return self.cw_extra_kwargs["etype"]

    @property
    def title(self):
        """Title of the collection, plural form of entity type."""
        return self._cw._("{}_plural").format(self.etype)

    def jsl_field(self, schema_role, **kwargs):
        """Return a JSON schema dict to model a collection of entities."""
        item_mapper = self._cw.vreg["mappers"].select(
            "jsonschema.item", self._cw
        )
        item_field = item_mapper.jsl_field(schema_role, **kwargs)
        return jsl.fields.ArrayField(title=self.title, items=item_field)

    def serialize(self, entities):
        """Return a list of collection item representing each entity in
        `entities`.
        """
        mapper = self._cw.vreg["mappers"].select("jsonschema.item", self._cw)
        return [mapper.serialize(entity) for entity in entities]


class RelatedCollectionMapper(EntityCollectionMapper):
    """Mapper for a collection of entities through a relation."""

    __select__ = match_kwargs("rtype", "role")

    @property
    def title(self):
        """Title of the collection, name of the relation."""
        rtype = self.cw_extra_kwargs["rtype"]
        role = self.cw_extra_kwargs["role"]
        return self._cw._(rtype if role == "subject" else rtype + "-object")


class CollectionItemMapper(JSONSchemaMapper):
    """Mapper for an item of a collection."""

    __regid__ = "jsonschema.item"
    __select__ = yes()

    def jsl_field(self, schema_role, **kwargs):
        """Return a jsl.fields.DictField with item description."""
        if schema_role == CREATION_ROLE:
            return jsl.fields.StringField(**kwargs)
        return jsl.fields.DictField(
            properties={
                "type": jsl.fields.StringField(),
                "id": jsl.fields.StringField(),
                "title": jsl.fields.StringField(),
            },
            **kwargs
        )

    @staticmethod
    def serialize(entity):
        """Return a dictionary with entity represented as a collection item."""
        return {
            "type": entity.cw_etype.lower(),
            "id": str(entity.eid),
            "title": entity.dc_title(),
        }


class BaseRelationMapper(JSONSchemaMapper):
    """Base abstract class to fill the gap between a yams relation and it's json
    schema mapping.

    They should be selected depending on the relation (`etype`, `rtype`, `role`
    and optionaly `target_types`). You may then get a Jsl field from them by
    calling the `jsl_field` method.
    """

    __regid__ = "jsonschema.relation"
    __select__ = match_kwargs("rtype", "role")
    __abstract__ = True

    @property
    def jsl_field_class(self):
        """The jsl.fields.BaseField child class
        to be returned by the factory."""
        raise NotImplementedError

    @property
    def etype(self):
        """The entity type bound to this mapper."""
        return _etype_from_context(self.cw_extra_kwargs)

    def __init__(self, _cw, **kwargs):
        #: relation type name
        self.rtype = kwargs.pop("rtype")
        #: role of `etype` in relation
        self.role = kwargs.pop("role")
        #: possible target types of the relation (empty for attribute relations)
        self.target_types = []
        rschema = _cw.vreg.schema[self.rtype]
        target_types = kwargs.pop("target_types", None)
        for target_eschema in sorted(rschema.targets(role=self.role)):
            if target_eschema.final:
                continue
            target_type = target_eschema.type
            if target_types is not None and target_type not in target_types:
                continue
            self.target_types.append(target_type)

        super(BaseRelationMapper, self).__init__(_cw, **kwargs)

    def __repr__(self):
        return (
            "<{0.__class__.__name__} etype={0.etype} rtype={0.rtype} "
            "role={0.role} target_types={0.target_types}>".format(self)
        )

    @cachedproperty
    def description(self):
        ntargets = len(self.target_types)
        if ntargets > 1:
            return None
        elif ntargets == 1:
            targettype = self.target_types[0]
        else:
            targettype = None
        eschema = self._cw.vreg.schema[self.etype]
        rdef = eschema.relation_definition(
            self.rtype, role=self.role, targettype=targettype
        )
        if rdef.description:
            return self._cw._(rdef.description)

    @cachedproperty
    def title(self):
        if self.role == "object":
            return self._cw._(self.rtype + "_object")
        return self._cw._(self.rtype)

    @cachedproperty
    def orm_rtype(self):
        return orm_rtype(self.rtype, self.role)

    def jsl_field(self, schema_role, **kwargs):
        """Return a JSL field built from `jsl_field_class`."""
        if "title" not in kwargs:
            kwargs["title"] = self.title
        if "description" not in kwargs and self.description is not None:
            kwargs["description"] = self.description
        kwargs.pop("module", None)
        return self.jsl_field_class(**kwargs)

    def values(self, entity, instance):
        """Return a dictionary holding deserialized value for this field,
        given the input entity (None on creation) and `instance` (JSON values
        as dictionary).

        If present, the field matching mapped `rtype` is removed from
        `instance` after processing.
        """
        raise NotImplementedError()

    def serialize(self, entity):
        """Return the serialized value for this field."""
        raise NotImplementedError()

    @staticmethod
    def _type(json_value):
        """Return properly typed value for use within a cubicweb's entity from
        given JSON value.

        Nothing to do by default.
        """
        return json_value


class AttributeMapper(BaseRelationMapper):
    """Base abstract class to map attribute relation."""

    __abstract__ = True
    __select__ = yams_final_rtype() & match_kwargs("etype")

    #: JSON Schema "format" keyword for semantic validation.
    format = None

    @property
    def attr(self):
        """Relation definition for bound attribute."""
        return self._cw.vreg.schema[self.etype].relation_definition(self.rtype)

    def add_constraint(self, cstr, jsl_attrs):
        serializer = self._cw.vreg["mappers"].select_or_none(
            "jsonschema.constraint",
            self._cw.vreg,
            self._cw._,
            self.etype,
            self.rtype,
            cstr,
            jsl_attrs,
        )
        if serializer is not None:
            new_attrs = serializer.todict()
            if new_attrs:
                jsl_attrs.update(new_attrs)
        elif not isinstance(cstr, StaticVocabularyConstraint):
            self.warning("JSL: ignored %s on %s", cstr.type(), self.attr)

    def jsl_field(self, schema_role, **kwargs):
        kwargs.setdefault("format", self.format)
        if hasattr(self, "title") and self.title:
            kwargs["title"] = self.title
        if schema_role in (CREATION_ROLE, EDITION_ROLE):
            if "required" not in kwargs and self.attr.cardinality[0] == "1":
                kwargs["required"] = True
            if "default" not in kwargs and self.attr.default is not None:
                kwargs["default"] = self.attr.default
            vocabulary_constraint = next(
                (
                    cstr
                    for cstr in self.attr.constraints
                    if isinstance(cstr, StaticVocabularyConstraint)
                ),
                None,
            )
            if vocabulary_constraint:
                # In presence of a vocabulary constraint, we wrap the field
                # into a oneOf field with a single-value 'enum', ignoring
                # other constraints.
                fields = []
                field_factory = super(AttributeMapper, self).jsl_field
                for v in sorted(vocabulary_constraint.vocabulary()):
                    kw = {"enum": [v], "title": self._cw._(v)}
                    fields.append(field_factory(schema_role, **kw))
                return jsl.fields.OneOfField(fields, **kwargs)
            for constraint in self.attr.constraints:
                self.add_constraint(constraint, kwargs)
        return super(AttributeMapper, self).jsl_field(schema_role, **kwargs)

    def values(self, entity, instance):
        """Return a dictionary holding deserialized value for mapped attribute.

        If mapped attribute is absent from `instance` and `entity` is not
        None, the default value for attribute is returned as value of the
        dictionnary.
        """
        try:
            value = instance.pop(self.rtype)
        except KeyError:
            if entity is None:
                return {}
            rschema = entity.e_schema.relation_definition(self.rtype, self.role)
            value = rschema.default
        else:
            value = self._type(value)
            if entity is not None and getattr(entity, self.orm_rtype) == value:
                # Do not trigger update, if the value has not changed.
                # This is useful for PUT requests for which all fields should
                # be present even if they are not changed. By skipping
                # unchanged value, we avoid possible security check that would
                # be meaningless since the value is not changed.
                return {}
        return {self.orm_rtype: value}

    def serialize(self, entity):
        value = getattr(entity, self.orm_rtype)
        if value is not None:
            return self._value(value)

    @staticmethod
    def _value(value):
        """Return the serializable value from attribute `value`."""
        return value


class ChoiceStringField(jsl.fields.StringField):
    """A string field with choices described in a oneOf array with items made
    of {'enum': [<choice value>], 'title': <choice title>} objects.
    """

    def __init__(self, choices, **kwargs):
        super(ChoiceStringField, self).__init__(**kwargs)
        self.choices = choices

    def get_definitions_and_schema(self, **kwargs):
        definitions, schema = super(
            ChoiceStringField, self
        ).get_definitions_and_schema(**kwargs)
        schema["oneOf"] = [
            {"enum": [value], "title": title} for value, title in self.choices
        ]
        return definitions, schema


class StringMapper(AttributeMapper):
    """Attribute mapper for Yams' String type."""

    __select__ = yams_match(target_types="String")
    jsl_field_class = jsl.fields.StringField
    _type = str


class FloatMapper(AttributeMapper):
    """Attribute mapper for Yams' Float type."""

    __select__ = yams_match(target_types="Float")
    jsl_field_class = jsl.fields.NumberField


class IntMapper(AttributeMapper):
    """Attribute mapper for Yams' Int type."""

    __select__ = yams_match(target_types="Int")
    jsl_field_class = jsl.fields.IntField


class BooleanMapper(AttributeMapper):
    """Attribute mapper for Yams' Boolean type."""

    __select__ = yams_match(target_types="Boolean")
    jsl_field_class = jsl.fields.BooleanField


class PasswordMapper(AttributeMapper):
    """Attribute mapper for Yams' Password type."""

    __select__ = yams_match(target_types="Password")
    jsl_field_class = jsl.fields.StringField
    #:
    format = "password"

    def jsl_field(self, schema_role, **kwargs):
        if schema_role == EDITION_ROLE:
            kwargs.setdefault("required", False)
        return super(PasswordMapper, self).jsl_field(schema_role, **kwargs)

    def values(self, entity, instance):
        password_changed = self.orm_rtype in instance
        values = super(PasswordMapper, self).values(entity, instance)
        if entity is not None and not password_changed:
            # We don't want the Password value to be reset if it has not
            # changed.
            del values[self.orm_rtype]
        return values

    def serialize(self, entity):
        return None

    @staticmethod
    def _type(json_value):
        """Return an encoded string suitable for Password type."""
        return json_value.encode("utf-8")


class DateMapper(StringMapper):
    """Attribute mapper for Yams' Date type."""

    __select__ = yams_match(target_types=("Date"))
    #:
    format = "date"

    @staticmethod
    def _type(value):
        """Return a datetime object parsed from ISO8601 `value` string."""
        return iso8601.parse_date(value)


class DatetimeMapper(DateMapper):
    """Attribute mapper for Yams' (TZ)Datetime type."""

    __select__ = yams_match(target_types=("Datetime", "TZDatetime"))
    #:
    format = "date-time"


class BytesMapper(StringMapper):
    """Attribute mapper for Yams' Bytes type."""

    __select__ = yams_match(target_types="Bytes")

    @staticmethod
    def _type(value):
        """Return a Binary containing `value`."""
        return Binary(value.encode("utf-8"))

    @staticmethod
    def _value(value):
        """Return a unicode string from Binary `value`."""
        return value.getvalue().decode("utf-8")


class CompoundMapper(JSONSchemaMapper):
    """Mapper for a "compound" field gathering Yams relations into a
    dedicated JSON "object" to be inserted in "definitions" key of the JSON
    Schema document.

    The compound "object" will appear in the main JSON Schema document's
    properties under the `name` class attribute (defaults to class name).
    """

    __regid__ = "jsonschema.object"
    __abstract__ = True
    __select__ = partial_yams_match()

    #: entity type holding this compound field.
    etype = None
    #: sequence of relations gathered in this compound field
    relations = ()
    #: title of the field
    title = None

    # sequence of 'relation/role' computed from 'relations'
    _relations = ()

    @classproperty
    def name(cls):
        """name of the property to be inserted in the main JSON Schema
        document (defaults to class name).
        """
        return str(cls.__name__)

    @classmethod
    def __registered__(cls, reg):
        if not cls.relations:
            raise ValueError(
                '{} is missing a "relations" class attribute'.format(cls)
            )
        # Compute relation/role pairs from relations
        relations = []
        for relinfo in cls.relations:
            try:
                rtype, role = relinfo
            except ValueError:
                rtype = relinfo
                role = "subject"
            relations.append((rtype, role))
        cls._relations = tuple(relations)
        # Check for name or relation duplication
        for obj in reg[cls.__regid__]:
            if obj == cls:
                continue
            if obj.etype == cls.etype:
                if obj.name == cls.name:
                    # Make sure 'name' is unique amongst objects with
                    # 'jsonschema.object' regid.
                    raise ValueError(
                        'a class with name "{}" is already '
                        "registered".format(cls.name)
                    )
                # Prevent duplicate mapping of the same etype/rtype.
                common_rtypes = set(obj._relations) & set(cls._relations)
                if common_rtypes:
                    raise ValueError(
                        'duplicate relation mapping for "{}": {}'.format(
                            cls.etype,
                            ", ".join("-".join(rel) for rel in common_rtypes),
                        )
                    )
        # Hide relations mapped to this document from etype JSON Schema.
        for rtype, role in cls._relations:
            if role == "object":
                jsonschema_section.tag_object_of(
                    ("*", rtype, cls.etype), "hidden"
                )
            else:
                jsonschema_section.tag_subject_of(
                    (cls.etype, rtype, "*"), "hidden"
                )
        return super(CompoundMapper, cls).__registered__(reg)

    @property
    def jsl_field_class(self):
        return partial(
            jsl.fields.DocumentField, as_ref=True, title=self._cw._(self.title)
        )

    def jsl_field(self, schema_role):
        attrs = {}
        for rtype, role in self._relations:
            rschema = self._cw.vreg.schema[rtype]
            target_types = {t.type for t in rschema.targets(self.etype, role)}
            mapper = self._cw.vreg["mappers"].select(
                "jsonschema.relation",
                self._cw,
                etype=self.etype,
                rtype=rtype,
                role=role,
                target_types=target_types,
            )
            attrs[mapper.orm_rtype] = mapper.jsl_field(schema_role)
        attrs["Options"] = default_options(self._cw, self.title)
        doc = type(str(self.title), (jsl.Document,), attrs)
        return self.jsl_field_class(doc)

    def values(self, entity, instance):
        assert (
            entity is None or entity.cw_etype == self.etype
        ), 'cannot get "values" for {} with {}'.format(entity, self)
        try:
            values = instance.pop(self.name)
        except KeyError:
            if entity is None:
                return {}
            # Continue with an empty "value" that would get filled by compound
            # relation mappers with default or None values.
            values = {}
        for rtype, role in self._relations:
            mapper = self._relation_mapper(rtype, role)
            values.update(mapper.values(entity, values))
        return values

    def serialize(self, entity):
        assert (
            entity.cw_etype == self.etype
        ), "cannot serialize {} with {}".format(entity, self)
        data = {}
        for rtype, role in self._relations:
            mapper = self._relation_mapper(rtype, role)
            value = mapper.serialize(entity)
            if value is not None:
                data[mapper.orm_rtype] = value
        return data

    def _relation_mapper(self, rtype, role="subject"):
        vreg = self._cw.vreg
        rschema = vreg.schema[rtype]
        target_types = {t.type for t in rschema.targets(self.etype, role)}
        return vreg["mappers"].select(
            "jsonschema.relation",
            self._cw,
            etype=self.etype,
            rtype=rtype,
            role=role,
            target_types=target_types,
        )


class _RelationMapper(BaseRelationMapper):
    """Abstract class for true relation (as opposed to attribute) mapper."""

    __abstract__ = True
    __select__ = ~yams_final_rtype()
    jsl_field_class = jsl.fields.ArrayField

    def jsl_field(self, schema_role, **kwargs):
        if "items" not in kwargs:
            targets = list(self.jsl_field_targets(schema_role))
            cardinalities = set([])
            rschema = self._cw.vreg.schema[self.rtype]
            for target_type in self.target_types:
                rdef = rschema.role_relation_definition(
                    self.etype, target_type, self.role
                )
                cardinalities.add(rdef.role_cardinality(self.role))
            if len(targets) > 1:
                kwargs["items"] = jsl.fields.OneOfField(targets)
            else:
                kwargs["items"] = targets[0]
            if schema_role in (CREATION_ROLE, EDITION_ROLE):
                cardinality = cardinalities.pop()
                if cardinalities:
                    raise BadSchemaDefinition(
                        "inconsistent cardinalities within {0} relation "
                        "definitions".format(self.rtype)
                    )
                if cardinality != "*":
                    if "required" not in kwargs and cardinality in "1+":
                        kwargs["required"] = True
                    if "min_items" not in kwargs:
                        kwargs["min_items"] = 0 if cardinality in "?" else 1
                    if "max_items" not in kwargs and cardinality in "?1":
                        kwargs["max_items"] = 1
        return super(_RelationMapper, self).jsl_field(schema_role, **kwargs)

    def jsl_field_targets(self, schema_role):
        """Return an iterator on jsl field objects to put in the 'items' key."""
        raise NotImplementedError()


class InlinedRelationMapper(_RelationMapper):
    """Map relation as 'inlined', i.e. the target of the relation is
    created/edited along with its original entity.
    """

    __select__ = (
        _RelationMapper.__select__
        & yams_component_target()
        & (match_kwargs("etype") | match_kwargs("entity"))
    )

    def jsl_field_targets(self, schema_role):
        for target_type in self.target_types:
            mapper = self._cw.vreg["mappers"].select(
                "jsonschema.entity",
                self._cw,
                etype=target_type,
                rtype=self.rtype,
                role=neg_role(self.role),
                target_types={self.etype},
            )
            target_document = mapper.jsl_document(schema_role)
            yield jsl.fields.DocumentField(target_document, as_ref=True)

    def values(self, entity, instance):
        # Would require knownledge of the target type from "instance",
        # but the generated JSON schema does not expose this yet.
        assert (
            len(self.target_types) == 1
        ), "cannot handle multiple target types yet: {}".format(
            self.target_types
        )
        target_type = self.target_types[0]
        try:
            values = instance.pop(self.rtype)
        except KeyError:
            if entity is None:
                return {}
            values = []
        if not isinstance(values, list):
            # TODO: rather jsonschema.validate()
            raise ValidationError(
                entity, {self.rtype: _("value should be an array")}
            )
        if entity is not None:
            # if entity already exists, delete entities related through
            # this mapped relation
            for linked_entity in getattr(entity, self.orm_rtype):
                if linked_entity.cw_etype in self.target_types:
                    linked_entity.cw_delete()
        target_mapper = self._cw.vreg["mappers"].select(
            "jsonschema.entity",
            self._cw,
            etype=target_type,
            rtype=self.rtype,
            role=neg_role(self.role),
            target_types={self.etype},
        )
        result = []
        for subinstance in values:
            subvalues = target_mapper.values(subinstance)
            result.append(self._cw.create_entity(target_type, **subvalues))
        return {self.orm_rtype: result}

    def serialize(self, entity):
        rset = entity.related(
            self.rtype, self.role, targettypes=tuple(self.target_types)
        )
        if not rset:
            return None

        def serialize(entity):
            mapper = self._cw.vreg["mappers"].select(
                "jsonschema.entity",
                self._cw,
                entity=entity,
                rtype=self.rtype,
                role=neg_role(self.role),
                target_types={self.etype},
            )
            return mapper.serialize()

        return [serialize(related) for related in rset.entities()]


class ETypeRelationMapper(_RelationMapper):
    """Map relation as 'generic', i.e. the target of the relation may be
    selected in preexisting possible targets.
    """

    __select__ = (
        _RelationMapper.__select__
        & ~yams_component_target()
        & match_kwargs("etype")
    )

    def relation_targets(self, schema_role):
        entity = self._cw.vreg["etypes"].etype_class(self.etype)(self._cw)
        potential_targets = []
        for target_type in self.target_types:
            potential_targets.extend(
                entity.unrelated(self.rtype, target_type, self.role).entities()
            )
        return potential_targets

    def jsl_field_targets(self, schema_role):
        choices = []
        for target in self.relation_targets(schema_role):
            choices.append((str(target.eid), target.dc_title()))
        yield ChoiceStringField(choices, type="string")

    def values(self, entity, instance):
        """
        Handling of reverse dependant entities creation is done by returning
        keys starting by __post_create. These keys contain a dictionary:
        {
            type: entity type used for creating the new entity
            values: a dictionary for attributes passed to this new entity
            relation: name of the relation added pointing to the main entity
        }
        These keys can also contain a list of such dictionaries.
        """
        try:
            values = instance.pop(self.rtype)
        except KeyError:
            return {}
        if entity is not None:
            entity.cw_set(**{self.orm_rtype: None})
        return {self.orm_rtype: [int(x) for x in values]}

    def serialize(self, entity):
        rset = entity.related(
            self.rtype, self.role, targettypes=tuple(self.target_types)
        )
        if not rset:
            return None
        return [str(x.eid) for x in rset.entities()]


class EntityRelationMapper(ETypeRelationMapper):
    """Map relation using existing relations target when role is view,
    fallback to ETypeMapper otherwise to display all possible relation
    targets.
    """

    __select__ = (
        _RelationMapper.__select__
        & ~yams_component_target()
        & match_kwargs("entity")
    )

    def relation_targets(self, schema_role):
        entity = self.cw_extra_kwargs["entity"]
        if schema_role == VIEW_ROLE:
            return entity.related(
                self.rtype, self.role, targettypes=tuple(self.target_types)
            ).entities()
        if schema_role == CREATION_ROLE:
            assert (
                len(self.target_types) == 1
            ), "cannot handle multiple target types in {} for {}".format(
                self, schema_role
            )
            targettype = self.target_types[0]
            return entity.unrelated(
                self.rtype,
                targettype,
                role=self.role,
            ).entities()
        return super(EntityRelationMapper, self).relation_targets(schema_role)



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""cubicweb-jsonschema hyper schema "link" components.

See http://json-schema.org/draft-06/links for the JSON schema of these
objects.
"""


from cubicweb.appobject import AppObject
from cubicweb.predicates import has_permission, match_kwargs

from .. import VIEW_ROLE, CREATION_ROLE, EDITION_ROLE


class Link(AppObject):
    """Abstract hyper schema link."""

    __registry__ = "links"

    def description_object(self, request, resource):
        """Return the link description object as a dict."""
        raise NotImplementedError()


class CollectionLink(Link):
    """Link for a collection."""

    __regid__ = "collection"
    __select__ = match_kwargs("title")
    rel = "collection"

    @property
    def title(self):
        return self.cw_extra_kwargs["title"]

    def description_object(self, request, resource):
        return {
            "href": request.resource_path(resource),
            "rel": self.rel,
            "targetSchema": {
                "$ref": request.resource_path(resource, "schema"),
            },
            "submissionSchema": {
                "$ref": request.resource_path(
                    resource, "schema", query={"role": CREATION_ROLE}
                ),
            },
            "title": self.title,
        }


class CollectionItemLink(Link):
    """Link for a collection item."""

    __regid__ = "collection-item"
    __select__ = match_kwargs("anchor")
    rel = "item"

    def description_object(self, request, resource):
        href = request.resource_path(resource) + "{id}"
        return {
            "href": href,
            "rel": self.rel,
            "anchor": self.cw_extra_kwargs["anchor"],
        }


class ETypeLink(CollectionLink):
    """Link for a collection of entities of the same type."""

    __regid__ = "collection.etype"
    __select__ = match_kwargs("etype")
    rel = "self"

    @property
    def title(self):
        etype = self.cw_extra_kwargs["etype"]
        return self._cw.__(etype + "_plural")


class EntityLink(Link):
    """Abstract hyper schema link suitable for entity JSON schema."""

    __abstract__ = True
    __select__ = match_kwargs("entity")

    @property
    def entity(self):
        return self.cw_extra_kwargs["entity"]


class EntitySelfLink(EntityLink):
    """Link for an entity as rel="self"."""

    __regid__ = "entity.self"
    __select__ = EntityLink.__select__ & has_permission("read")

    def description_object(self, request, resource):
        entity = self.entity
        title = self._cw._("{0} #{1}").format(
            self._cw.__(entity.cw_etype), entity.eid
        )
        return {
            "href": request.resource_path(resource),
            "rel": "self",
            "title": title,
            "submissionSchema": {
                "$ref": request.resource_path(
                    resource, "schema", query={"role": EDITION_ROLE}
                ),
            },
            "targetSchema": {
                "$ref": request.resource_path(
                    resource, "schema", query={"role": VIEW_ROLE}
                ),
            },
        }


class EntityParentLink(EntityLink):
    """Link to the parent resource of an entity (i.e. the entity type)."""

    __regid__ = "entity.type"
    rel = "up"

    def description_object(self, request, resource):
        return {
            "href": request.resource_path(resource.__parent__),
            "rel": self.rel,
            "title": self._cw.__("{0}_plural").format(self.entity.cw_etype),
            "targetSchema": {
                "$ref": request.resource_path(resource.__parent__, "schema"),
            },
        }


class EntityCollectionLink(EntityParentLink):
    """Link for an entity as item of a collection."""

    __regid__ = "entity.collection"
    rel = "collection"


class EntityRelatedLink(EntityLink):
    """Link to (`rtype`, `role`) relation from `entity` as `role`."""

    __regid__ = "entity.related"
    __select__ = match_kwargs("entity", "rtype", "role")

    def description_object(self, request, resource):
        assert resource.__class__.__name__ == "EntityResource"  # XXX debug
        rtype = self.cw_extra_kwargs["rtype"]
        role = self.cw_extra_kwargs["role"]
        collection_mapper = self._cw.vreg["mappers"].select(
            "jsonschema.collection", self._cw, rtype=rtype, role=role
        )
        title = collection_mapper.title
        return {
            "href": request.resource_path(resource[rtype]),
            "rel": "related",
            "title": title,
        }


class RelatedCollectionLink(CollectionLink):
    """rel="self" link for collection of related entities."""

    __regid__ = "related-collection"
    rel = "self"

    def description_object(self, request, resource):
        ldo = super(RelatedCollectionLink, self).description_object(
            request, resource
        )
        ldo["targetSchema"]["$ref"] = request.resource_path(
            resource, "schema", query={"role": VIEW_ROLE}
        )
        return ldo



from yams.buildobjs import (
    EntityType,
    Boolean,
    Bytes,
    ComputedRelation,
    Date,
    Float,
    Int,
    RelationDefinition,
    String,
    RichString,
)
from cubicweb.schema import (
    ERQLExpression,
    RRQLExpression,
)


class Photo(EntityType):
    data = Bytes(required=True)
    media_type = String(
        vocabulary=["png", "jpeg"], required=True, default="png"
    )
    exposure_time = Float()
    flash = Boolean(required=True, default=False)
    maker_note = Bytes()


class Thumbnail(EntityType):
    data = Bytes(required=True)


class thumbnail(RelationDefinition):
    subject = "Photo"
    object = "Thumbnail"
    cardinality = "?1"
    composite = "subject"
    inlined = True


class picture(RelationDefinition):
    subject = "CWUser"
    object = "Photo"
    cardinality = "*?"
    composite = "subject"


class EmailAlias(EntityType):
    pass


class use_email(RelationDefinition):
    subject = "CWUser"
    object = "EmailAlias"
    cardinality = "*1"
    composite = "subject"


class Book(EntityType):
    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": ("managers", "users"),
        "update": (ERQLExpression("X owned_by U"),),
        "delete": ("managers", "users", "guests"),
    }
    title = String(required=True)
    publication_date = Date()


class Author(EntityType):
    name = String(required=True)


class author(RelationDefinition):
    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": (RRQLExpression("U has_update_permission S"),),
        "delete": (RRQLExpression("U has_update_permission S"),),
    }
    subject = "Book"
    object = "Author"
    cardinality = "?*"


class publications(ComputedRelation):
    rule = "O author S"


class Topic(EntityType):
    name = String(
        required=True,
        __permissions__={
            "read": ("managers", "users", "guests"),
            "add": ("managers", "users"),
            "update": (
                "managers",
                ERQLExpression("NOT EXISTS(O topics X)"),
            ),
        },
    )


class topics(RelationDefinition):
    subject = "Book"
    object = "Topic"



# copyright 2016-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""cubicweb-jsonschema Pyramid views for the JSON API of entities."""

from pyramid import httpexceptions
from pyramid.renderers import render
from pyramid.traversal import (
    find_resource,
    find_root,
)
from pyramid.view import view_config

from cubicweb import (
    neg_role,
    ValidationError,
)
from ..resources import (
    EntityResource,
    ETypeResource,
    RelatedEntitiesResource,
    RelatedEntityResource,
    RelationshipResource,
    RootResource,
)
from . import (
    describedby,
    entity_from_context,
    json_config,
    json_problem,
    jsonschema_adapter,
    LOG,
)


@view_config(
    route_name="cubicweb-jsonschema.entities",
    context=RootResource,
    request_method="GET",
    decorator=[describedby],
)
def get_root(context, request):
    """Site root, no data."""
    response = httpexceptions.HTTPNoContent()
    response.allow = ["GET"]
    return response


@json_config(
    route_name="cubicweb-jsonschema.entities",
    context=ETypeResource,
    request_method="GET",
    decorator=[describedby],
)
def get_entities(context, request):
    """Render multiple entities in JSON format."""
    adapter = jsonschema_adapter(request.cw_request, etype=context.etype)
    vreg = request.registry["cubicweb.registry"]
    if vreg.schema[context.etype].has_perm(request.cw_cnx, "add"):
        request.response.allow = ["GET", "POST"]
    else:
        request.response.allow = ["GET"]
    return adapter.serialize(context.rset.entities())


def allow_for_entity(entity) -> list:
    """Return a list of HTTP verbs that are Allow-ed for `entity`."""
    PERM2VERB = [
        ("read", "GET"),
        ("update", "PUT"),
        ("delete", "DELETE"),
    ]
    return [v for p, v in PERM2VERB if entity.cw_has_perm(p)]


@json_config(
    route_name="cubicweb-jsonschema.entities",
    context=EntityResource,
    request_method="GET",
    decorator=[entity_from_context, describedby],
)
def get_entity(context, request):
    """Render a single entity in JSON format."""
    entity = context.entity
    request.response.allow = allow_for_entity(entity)
    return entity.cw_adapt_to("IJSONSchema").serialize()


@json_config(
    route_name="cubicweb-jsonschema.entities",
    context=ETypeResource,
    request_method="POST",
)
def create_entity(context, request):
    """Create a new entity from JSON data."""
    # TODO In case of validation errors, it'd be better to give a JSON Schema
    # entry as a "pointer", would require selection context to be an
    # ETypeSchemaResource.
    adapter = jsonschema_adapter(request.cw_request, etype=context.etype)
    instance = request.json_body
    entity = adapter.create_entity(instance)
    request.cw_cnx.commit()
    LOG.info("created %s", entity)

    value = entity.cw_adapt_to("IJSONSchema").serialize()
    entity_resource = context[entity.eid]
    location = request.resource_url(entity_resource)
    raise httpexceptions.HTTPCreated(
        location=location,
        content_type="application/json; charset=UTF-8",
        body=render("json", value),
    )


@json_config(
    route_name="cubicweb-jsonschema.entities",
    context=EntityResource,
    request_method="PUT",
    decorator=[entity_from_context],
)
def update_entity(context, request):
    """Update an entity from JSON data."""
    entity = context.entity
    instance = request.json_body
    adapted = entity.cw_adapt_to("IJSONSchema")
    adapted.edit_entity(instance)
    request.cw_cnx.commit()
    LOG.info('edited attributes "%s" of %s', ", ".join(instance), entity)
    value = adapted.serialize()
    raise httpexceptions.HTTPOk(
        location=request.resource_url(context),
        content_type="application/json; charset=UTF-8",
        body=render("json", value),
    )


@view_config(
    route_name="delete_entity",
    context=EntityResource,
    request_method="DELETE",
    decorator=[entity_from_context],
)
def delete_entity(context, request):
    """Delete an entity."""
    entity = context.entity
    entity.cw_delete()
    request.cw_cnx.commit()
    LOG.info("deleted %s", entity)
    raise httpexceptions.HTTPNoContent()


def allow_for_entity_relation(entity, rtype, role: str) -> list:
    eschema = entity.e_schema
    rdef = eschema.relation_definition(rtype, role)
    PERM2VERB = [
        ("read", "GET"),
        ("add", "POST"),
    ]
    if role == "subject":
        kwargs = {"fromeid": entity.eid}
    else:
        kwargs = {"toeid": entity.eid}
    return [v for p, v in PERM2VERB if rdef.has_perm(entity._cw, p, **kwargs)]


@json_config(
    route_name="cubicweb-jsonschema.entities",
    context=RelatedEntitiesResource,
    request_method="GET",
    decorator=[describedby],
)
def get_related_entities(context, request):
    """Return a JSON document of entities target of a relationships."""
    vreg = request.registry["cubicweb.registry"]
    mapper = vreg["mappers"].select(
        "jsonschema.collection",
        request.cw_request,
        rtype=context.rtype,
        role=context.role,
    )
    entity = context.__parent__.rset.one()
    request.response.allow = allow_for_entity_relation(
        entity, context.rtype, context.role
    )
    return mapper.serialize(context.rset.entities())


@json_config(
    route_name="cubicweb-jsonschema.entities",
    context=RelatedEntitiesResource,
    request_method="POST",
)
def post_related_entities(context, request):
    """Insert relationships between entity bound to context and targets from
    request's JSON body.
    """
    try:
        targets_eid = [int(x) for x in request.json_body]
    except ValueError:
        raise httpexceptions.HTTPBadRequest("invalid target identifier(s)")
    entity = context.__parent__.rset.one()
    rtype = context.rtype
    if context.role != "subject":
        raise httpexceptions.HTTPNotFound(
            "relation {}-subject not found on {}".format(rtype, entity.cw_etype)
        )
    # TODO: move this to an IJSONSchema adapter
    entity.cw_set(**{rtype: targets_eid})
    request.cw_cnx.commit()
    return httpexceptions.HTTPNoContent()


@json_config(
    route_name="cubicweb-jsonschema.entities",
    context=RelatedEntityResource,
    request_method="GET",
    decorator=[describedby],
)
def get_related_entity(context, request):
    """Return a JSON document of an entity target of a relationship."""
    entity = context.entity
    adapted = jsonschema_adapter(
        request.cw_cnx,
        entity=entity,
        rtype=context.rtype,
        role=neg_role(context.role),
    )
    request.response.allow = allow_for_entity(entity)
    return adapted.serialize()


@json_config(
    route_name="cubicweb-jsonschema.entities",
    context=RelatedEntityResource,
    request_method="PUT",
)
def update_related_entity(context, request):
    """Update a related entity from JSON data."""
    instance = request.json_body
    entity = context.entity
    adapted = jsonschema_adapter(
        request.cw_cnx,
        entity=entity,
        rtype=context.rtype,
        role=neg_role(context.role),
    )
    adapted.edit_entity(instance)
    request.cw_cnx.commit()
    LOG.info('edited attributes "%s" of %s', ", ".join(instance), entity)
    value = adapted.serialize()
    raise httpexceptions.HTTPOk(
        location=request.resource_url(context),
        content_type="application/json; charset=UTF-8",
        body=render("json", value),
    )


@view_config(
    route_name="delete_entity",
    context=RelatedEntityResource,
    request_method="DELETE",
)
def delete_relation(context, request):
    """Delete a relationship."""
    assert context.role == "subject", "can only delete subject relation"
    subj = context.__parent__.__parent__.rset.one()
    obj = context.entity
    request.cw_cnx.execute(
        "DELETE S {rtype} O WHERE S eid %(s)s, O eid %(o)s".format(
            rtype=context.rtype
        ),
        {"s": subj.eid, "o": obj.eid},
    )
    request.cw_cnx.commit()
    LOG.info("deleted <%s %s %s>", subj, context.rtype, obj)
    raise httpexceptions.HTTPNoContent()


@json_config(
    route_name="cubicweb-jsonschema.entities",
    context=RelationshipResource,
    request_method="POST",
    decorator=[entity_from_context],
)
def create_related_entity(context, request):
    """Create an entity with a relation set."""
    rtype, role, target_type = context.rtype, context.role, context.target_type
    adapter = jsonschema_adapter(
        request.cw_request, etype=target_type, rtype=rtype, role=neg_role(role)
    )
    instance = request.json_body
    target = context.entity
    related = adapter.create_entity(instance, target)
    request.cw_cnx.commit()
    LOG.info(
        "created %s target of %s-%s for %s",
        related,
        rtype,
        role,
        context.entity,
    )
    related_adapted = jsonschema_adapter(
        request.cw_request, entity=related, rtype=rtype, role=neg_role(role)
    )
    value = related_adapted.serialize()
    root_resource = find_root(context)
    entity_resource = find_resource(
        root_resource, (related.cw_etype, related.eid)
    )
    location = request.resource_url(entity_resource)
    raise httpexceptions.HTTPCreated(
        location=location,
        content_type="application/json; charset=UTF-8",
        body=render("json", value),
    )


@json_config(
    route_name="cubicweb-jsonschema.entities",
    context=ValidationError,
)
def validation_failed(exc, request):
    """Exception view for ValidationError on JSON request."""
    LOG.info("%s encountered during processing of %s", exc, request)
    _ = request.cw_request._
    request.cw_cnx.rollback()
    exc.translate(_)
    invalid_params = []
    detail = []
    for rolename, value in exc.errors.items():
        if rolename:
            invalid_params.append({"name": rolename, "reason": value})
        else:
            detail.append(value)
    kwargs = {}
    if detail:
        kwargs["detail"] = "\n".join(detail)
    if invalid_params:
        kwargs["invalid-params"] = invalid_params
    return json_problem(status=422, **kwargs)


@json_config(
    route_name="cubicweb-jsonschema.entities",
    context=Exception,
)
def generic_error(exc, request):
    """Generic exception handler for exception (usually during "edition"
    operations).

    It will return a jsonapi-formatted error response from exception found in
    context.
    """
    # Because this handler is wired on the "Exception" context, any error will
    # be catched, including the standard pyramid return exceptions.
    # In this case, just return it
    if isinstance(exc, httpexceptions.HTTPException):
        return exc
    if request.cw_cnx is not None:
        request.cw_cnx.rollback()
    LOG.exception("exception occurred while processing %s", request)
    kwargs = {}
    detail = str(exc)
    if detail:
        kwargs["detail"] = detail
    return json_problem(status=500, **kwargs)


def includeme(config) -> None:
    config.include("cubicweb.pyramid.predicates")
    config.include("..predicates")
    config.add_route(
        "delete_entity",
        "/jsonschema/{etype}/*traverse",
        factory=ETypeResource.from_match("etype"),
        request_method="DELETE",
        match_is_etype="etype",
    )
    config.add_route(
        "cubicweb-jsonschema.entities",
        "/jsonschema/*traverse",
        factory=RootResource,
        strict_accept="application/json",
    )
    config.scan(__name__)



"""cubicweb-jsonschema application package

JSON Schema for CubicWeb
"""

VIEW_ROLE = "view"
CREATION_ROLE = "creation"
EDITION_ROLE = "edition"

JSONSCHEMA_MEDIA_TYPE = "application/schema+json"


def orm_rtype(rtype, role):
    if role == "object":
        return "reverse_" + rtype
    return rtype



# copyright 2016-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""cubicweb-jsonschema Yams to jsl  unit tests."""

from cubicweb.devtools.testlib import CubicWebTC

from cubicweb_jsonschema.views.y2jsl import y2j_etype


class CWYams2JsonTC(CubicWebTC):

    def rtype(self, etype, rtype):
        return self.vreg.schema[etype].relation_definition(rtype)

    def test_auto_attributes(self):
        with self.admin_access.cnx() as cnx:

            @y2j_etype(self.vreg, cnx)
            class CWUser:
                pass

        j_schema = CWUser.get_schema(ordered=True)
        j_props = j_schema["properties"]
        # Check attribute list
        self.assertEqual(
            [
                "__etype__",
                "__eid__",
                "login",
                "upassword",
                "firstname",
                "surname",
                "last_login_time",
            ],
            list(j_props),
        )
        self.assertEqual(
            ["__etype__", "__eid__", "login", "upassword"], j_schema["required"]
        )
        # Check type and description
        login = self.rtype("CWUser", "login")
        self.assertEqual(login.description, j_props["login"].get("description"))
        self.assertEqual("string", j_props["login"].get("type"))
        # Check format
        self.assertEqual("string", j_props["last_login_time"].get("type"))
        self.assertEqual("date-time", j_props["last_login_time"].get("format"))
        # Check simple constraint
        firstname = self.rtype("CWUser", "firstname")
        self.assertEqual(
            firstname.constraint_by_type("SizeConstraint").max,
            j_props["firstname"].get("maxLength"),
        )

    def test_auto_relations(self):
        with self.admin_access.cnx() as cnx:

            @y2j_etype(self.vreg, cnx)
            class Workflow:
                __filter_in__ = ("name", "description")

            @y2j_etype(self.vreg, cnx)
            class State:
                pass

            userworkflow_eid = cnx.execute(
                'Any X WHERE X workflow_of ET, ET name "CWUser"'
            )[0][0]

        j_schema = State.get_schema(ordered=True)
        state_props = j_schema["properties"]

        self.assertEqual(
            ["__etype__", "__eid__", "name", "description", "state_of"],
            list(state_props),
        )
        expected_items = {
            "type": "string",
            "oneOf": [
                {
                    "enum": [str(userworkflow_eid)],
                    "title": "default user workflow",
                }
            ],
        }
        self.assertEqual(state_props["state_of"]["items"], expected_items)
        self.assertEqual(state_props["state_of"]["minItems"], 1)
        self.assertEqual(state_props["state_of"]["maxItems"], 1)

    def test_reverse_relation(self):
        with self.admin_access.cnx() as cnx:

            @y2j_etype(self.vreg, cnx)
            class CWUser:
                pass

            @y2j_etype(self.vreg, cnx)
            class CWGroup:
                __filter_in__ = ("reverse_in_group",)

            users = [
                (user.eid, user.dc_title())
                for user in cnx.find("CWUser").entities()
            ]

        j_schema = CWGroup.get_schema(ordered=True)
        properties = j_schema["properties"]
        self.assertIn("in_group", properties)
        expected_items = {
            "type": "string",
            "oneOf": [
                {"enum": [str(eid)], "title": title} for eid, title in users
            ],
        }
        self.assertEqual(properties["in_group"]["items"], expected_items)

    def test_custom_attribute(self):
        with self.admin_access.cnx() as cnx:

            @y2j_etype(self.vreg, cnx)
            class CWUser:
                upassword = {"min_length": 7}

        j_schema = CWUser.get_schema(ordered=True)
        j_props = j_schema["properties"]
        self.assertIn("minLength", j_props["upassword"])
        self.assertEqual(7, j_props["upassword"]["minLength"])

    def test_custom_relation(self):
        with self.admin_access.cnx() as cnx:

            @y2j_etype(self.vreg, cnx)
            class CWGroup:
                pass

            @y2j_etype(self.vreg, cnx)
            class CWUser:
                in_group = {"min_items": 1}

        j_schema = CWUser.get_schema(ordered=True)
        user_props = j_schema["properties"]
        self.assertIn("in_group", user_props)
        self.assertEqual(1, user_props["in_group"]["minItems"])


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""cubicweb-jsonschema views tests"""


from cubicweb.devtools.testlib import CubicWebTC

from cubicweb_jsonschema.views import (
    jsonschema_section,
)


class JSONSchemaRelationTagsTC(CubicWebTC):

    def test_init(self):
        tagdefs = jsonschema_section._tagdefs
        for tagged, expected_section in (
            # meta rtype
            (("Book", "eid", "Int", "subject"), "hidden"),
            # ignored rtypes
            (("Book", "created_by", "CWUser", "subject"), "hidden"),
            # attribute
            (("Book", "title", "String", "subject"), "inlined"),
            # inlined relation, role-cardinality '1' or '+'
            (("CWUser", "in_group", "CWGroup", "subject"), "inlined"),
            # inlined relation, composite
            (("CWUser", "use_email", "EmailAddress", "object"), "inlined"),
            # plain relation
            (("Book", "topics", "Topic", "subject"), "related"),
            (("Book", "topics", "Topic", "object"), "related"),
            # computed relation
            (("Author", "publications", "Book", "object"), "related"),
        ):
            with self.subTest(tagged=tagged):
                self.assertEqual(tagdefs[tagged], expected_section)

    def fake_entity(self, cnx, etype):
        return self.vreg["etypes"].etype_class(etype)(cnx)

    def rtag(self, cnx, entity):
        return self.vreg["uicfg"].select("jsonschema", cnx, entity=entity)

    def test_relations_by_section_password(self):
        """upassword for CWUser is listed for "add" action, not for "read"."""
        with self.admin_access.cnx() as cnx:
            entity = self.fake_entity(cnx, "CWUser")
            rsection = self.rtag(cnx, entity)
            for rtype, _, _ in rsection.relations_by_section(
                entity, "inlined", "read"
            ):
                if rtype == "upassword":
                    self.fail('"upassword" unexpectedly found')
            add_rtypes = [
                rtype
                for rtype, _, _ in rsection.relations_by_section(
                    entity, "inlined", "add"
                )
            ]
            self.assertIn("upassword", add_rtypes)

    def test_relations_by_section_ignored(self):
        """meta rtypes are not listed."""
        with self.admin_access.cnx() as cnx:
            entity = self.fake_entity(cnx, "Book")
            rsection = self.rtag(cnx, entity)
            read_rtypes = [
                rtype
                for rtype, _, _ in rsection.relations_by_section(
                    entity, "inlined", "read"
                )
            ]
            self.assertNotIn("eid", read_rtypes)
            add_rtypes = [
                rtype
                for rtype, _, _ in rsection.relations_by_section(
                    entity, "related", "add"
                )
            ]
            self.assertNotIn("custom_workflow", add_rtypes)

    def test_relations_by_section_inlined(self):
        with self.admin_access.cnx() as cnx:
            entity = self.fake_entity(cnx, "Book")
            rsection = self.rtag(cnx, entity)
            rtypes = [
                rtype
                for rtype, _, _ in rsection.relations_by_section(
                    entity, "inlined", "read"
                )
            ]
            self.assertCountEqual(["title", "publication_date"], rtypes)

    def test_relations_by_section_related(self):
        with self.admin_access.cnx() as cnx:
            entity = self.fake_entity(cnx, "Book")
            rsection = self.rtag(cnx, entity)
            rtypes = [
                rtype
                for rtype, _, _ in rsection.relations_by_section(
                    entity, "related", "read"
                )
            ]
            self.assertCountEqual(["author", "publications", "topics"], rtypes)

    def test_relations_by_section_inlined_permissions(self):
        with self.admin_access.cnx() as cnx:
            cnx.create_entity("Book", title="t")
            self.create_user(cnx, "bob")
            cnx.commit()
        with self.new_access("bob").cnx() as cnx:
            entity = cnx.find("Book").one()
            rsection = self.rtag(cnx, entity)
            read_rtypes = [
                rtype
                for rtype, _, _ in rsection.relations_by_section(
                    entity, "inlined", "read"
                )
            ]
            self.assertCountEqual(["title", "publication_date"], read_rtypes)
            update_rtypes = [
                rtype
                for rtype, _, _ in rsection.relations_by_section(
                    entity, "inlined", "update"
                )
            ]
            self.assertCountEqual([], update_rtypes)

    def test_relations_by_section_related_permissions(self):
        with self.admin_access.cnx() as cnx:
            author = cnx.create_entity("Author", name="a")
            cnx.create_entity("Book", title="t", author=author)
            self.create_user(cnx, "bob")
            cnx.commit()
        with self.new_access("bob").cnx() as cnx:
            entity = cnx.find("Book").one()
            rsection = self.rtag(cnx, entity)
            update_rtypes = [
                rtype
                for rtype, _, _ in rsection.relations_by_section(
                    entity, "related", "add"
                )
            ]
            self.assertCountEqual(["topics"], update_rtypes)


if __name__ == "__main__":
    import unittest

    unittest.main()



try:
    from cubicweb_web.views import uicfg
except ModuleNotFoundError:
    from cubicweb.web.views import uicfg  # TODO remove with CubicWeb 3 support

from cubicweb_jsonschema.views import jsonschema_section


uicfg.indexview_etype_section["Photo"] = "subobject"
uicfg.indexview_etype_section["Topic"] = "subobject"

jsonschema_section.tag_attribute(("CWUser", "lastname"), "hidden")
jsonschema_section.tag_subject_of(
    ("CWUser", "use_email", "EmailAddress"), "inlined"
)
jsonschema_section.tag_subject_of(
    ("CWUser", "use_email", "EmailAlias"), "hidden"
)
jsonschema_section.tag_object_of(("*", "use_email", "EmailAlias"), "hidden")
jsonschema_section.tag_subject_of(("CWUser", "picture", "*"), "inlined")
jsonschema_section.tag_object_of(("*", "picture", "Photo"), "hidden")
jsonschema_section.tag_subject_of(("Photo", "thumbnail", "*"), "inlined")

# we don't want this to interfere with entity creation/edition tests
jsonschema_section.tag_subject_of(
    ("CWUser", "use_email", "EmailAlias"), "hidden"
)



# copyright 2016-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""cubicweb-jsonschema Pyramid views for JSON Schema endpoints."""

from pyramid import httpexceptions
from pyramid.view import view_config

from cubicweb import neg_role

try:
    from cubicweb_web.views import uicfg
except ModuleNotFoundError:
    from cubicweb.web.views import uicfg  # TODO remove with CubicWeb 3 support

from .. import (
    CREATION_ROLE,
    EDITION_ROLE,
    JSONSCHEMA_MEDIA_TYPE,
    VIEW_ROLE,
    resources,
)
from . import (
    describes,
)


def jsonschema_config(**settings):
    for name, value in [
        ("name", "schema"),
        ("route_name", "cubicweb-jsonschema.schema"),
        ("accept", JSONSCHEMA_MEDIA_TYPE),
        ("renderer", "json"),
        ("request_method", "GET"),
    ]:
        settings.setdefault(name, value)
    return view_config(**settings)


def links_sortkey(link):
    """key function for sorting a list of link description object."""
    return link["href"], link["rel"]


@jsonschema_config(
    context=resources.RootResource,
    decorator=[describes],
)
def application_schema(context, request):
    """Schema view for the application."""
    vreg = request.registry["cubicweb.registry"]

    def links():
        for eschema in vreg.schema.entities():
            if uicfg.indexview_etype_section.get(eschema) != "application":
                continue
            etype = eschema.type
            etype_resource = context[etype]
            title = etype + "_plural"
            link = vreg["links"].select(
                "collection", request.cw_request, title=title
            )
            yield link.description_object(request, etype_resource)

    return {
        "title": request.cw_cnx.property_value("ui.site-title"),
        "type": "null",
        "links": sorted(links(), key=links_sortkey),
    }


@jsonschema_config(
    context=resources.ETypeResource,
    request_param="role",
)
def etype_role_schema(context, request):
    """Schema view for an entity type with specified role."""
    req = request.cw_request
    adapted = req.vreg["adapters"].select(
        "IJSONSchema", req, etype=context.etype
    )
    role = request.params["role"].lower()
    if role == VIEW_ROLE:
        return adapted.view_schema(ordered=True)
    elif role == CREATION_ROLE:
        return adapted.creation_schema(ordered=True)
    else:
        raise httpexceptions.HTTPBadRequest("invalid role: {0}".format(role))


@jsonschema_config(
    context=resources.ETypeResource,
    decorator=[describes],
)
def etype_schema(context, request):
    """Schema view for an entity type returning the complete hyper schema."""
    req = request.cw_request
    vreg = request.registry["cubicweb.registry"]
    collection_mapper = vreg["mappers"].select(
        "jsonschema.collection", req, etype=context.etype
    )
    schema = collection_mapper.jsl_field(VIEW_ROLE).get_schema(ordered=True)
    # Insert "rel": "item" link with "items" of the collection array with an
    # "anchor" pointing at the main instance (i.e. the collection).
    item_link = vreg["links"].select("collection-item", req, anchor="#")
    schema["items"]["links"] = [item_link.description_object(request, context)]
    schema["links"] = sorted(
        [
            link.description_object(request, context)
            for link in vreg["links"].possible_objects(
                req, etype=context.etype, rset=context.rset
            )
        ],
        key=links_sortkey,
    )
    return schema


@jsonschema_config(
    context=resources.EntityResource,
    request_param="role",
)
def entity_role_schema(context, request):
    """Schema view for a live entity with specified role."""
    entity = context.rset.one()  # May raise HTTPNotFound.
    adapted = entity.cw_adapt_to("IJSONSchema")
    role = request.params["role"].lower()
    if role == VIEW_ROLE:
        return adapted.view_schema(ordered=True)
    elif role == EDITION_ROLE:
        return adapted.edition_schema(ordered=True)
    else:
        raise httpexceptions.HTTPBadRequest("invalid role: {0}".format(role))


@jsonschema_config(
    context=resources.EntityResource,
    decorator=[describes],
)
def entity_schema(context, request):
    """Schema view for a live entity returning the complete hyper schema."""
    req = request.cw_request
    entity = context.rset.one()  # May raise HTTPNotFound.
    adapted = entity.cw_adapt_to("IJSONSchema")
    schema = adapted.view_schema(ordered=True)
    vreg = request.registry["cubicweb.registry"]
    linksreg = vreg["links"]
    links = [
        link.description_object(request, context)
        for link in linksreg.possible_objects(req, entity=entity)
    ]
    for rtype, role, __ in adapted.relations():
        if role != "subject":
            continue
        links.append(
            linksreg.select(
                "entity.related",
                req,
                entity=entity,
                rtype=rtype,
                role=role,
            ).description_object(request, context)
        )
    schema["links"] = sorted(links, key=links_sortkey)
    return schema


@jsonschema_config(
    context=resources.RelatedEntitiesResource,
    decorator=[describes],
)
def related_entities_schema(context, request):
    """Schema view for the collection of targets of a relation."""
    req = request.cw_request
    rtype, role = context.rtype, context.role
    vreg = request.registry["cubicweb.registry"]
    collection_mapper = vreg["mappers"].select(
        "jsonschema.collection", req, rtype=rtype, role=role
    )
    schema = collection_mapper.jsl_field(VIEW_ROLE).get_schema(ordered=True)
    # Insert "rel": "item" link with "items" of the collection array with an
    # "anchor" pointing at the main instance (i.e. the collection).
    item_link = vreg["links"].select("collection-item", req, anchor="#")
    schema["items"]["links"] = [item_link.description_object(request, context)]
    relatedcollection_link = vreg["links"].select(
        "related-collection", req, title=collection_mapper.title
    )
    schema["links"] = [
        relatedcollection_link.description_object(request, context),
    ]
    return schema


@jsonschema_config(
    context=resources.RelatedEntitiesResource,
    request_param="role",
)
def related_entities_role_schema(context, request):
    """Schema view for targets of a relation with specified 'role'."""
    req = request.cw_request
    vreg = request.registry["cubicweb.registry"]
    entity = context.__parent__.rset.one()
    mapper = vreg["mappers"].select(
        "jsonschema.relation",
        req,
        entity=entity,
        rtype=context.rtype,
        role=context.role,
    )
    schema_role = request.params["role"].lower()
    if schema_role not in (VIEW_ROLE, CREATION_ROLE):
        raise httpexceptions.HTTPBadRequest(
            "invalid role: {0}".format(schema_role)
        )
    return mapper.jsl_field(schema_role).get_schema(ordered=True)


@jsonschema_config(
    context=resources.RelatedEntityResource,
    decorator=[describes],
)
def related_entity_schema(context, request):
    """Schema view for a target entity of a relation."""
    req = request.cw_request
    entity = context.entity
    vreg = request.registry["cubicweb.registry"]
    mapper = vreg["mappers"].select(
        "jsonschema.entity",
        req,
        entity=entity,
        rtype=context.__parent__.rtype,
        role=context.__parent__.role,
    )
    schema = mapper.jsl_document(VIEW_ROLE).get_schema(ordered=True)
    linksreg = vreg["links"]
    links = [
        link.description_object(request, context)
        for link in linksreg.possible_objects(req, entity=entity)
    ]
    schema["links"] = sorted(links, key=links_sortkey)
    return schema


@jsonschema_config(
    context=resources.RelatedEntityResource,
    request_param="role",
)
def related_entity_role_schema(context, request):
    """Schema view for a target entity of a relation with specified 'role'."""
    req = request.cw_request
    entity = context.entity
    vreg = request.registry["cubicweb.registry"]
    mapper = vreg["mappers"].select(
        "jsonschema.entity",
        req,
        entity=entity,
        rtype=context.__parent__.rtype,
        role=context.__parent__.role,
    )
    schema_role = request.params["role"].lower()
    if schema_role not in (VIEW_ROLE, EDITION_ROLE):
        raise httpexceptions.HTTPBadRequest(
            "invalid role: {0}".format(schema_role)
        )
    return mapper.jsl_document(schema_role).get_schema(ordered=True)


@jsonschema_config(
    context=resources.RelationshipResource,
    request_param="role",
)
def relationship_schema(context, request):
    rtype, role, target_type = context.rtype, context.role, context.target_type
    adapter = request.cw_request.vreg["adapters"].select(
        "IJSONSchema",
        request.cw_request,
        etype=target_type,
        rtype=rtype,
        role=neg_role(role),
    )
    schema_role = request.params["role"].lower()
    if schema_role == CREATION_ROLE:
        return adapter.creation_schema(ordered=True)
    elif schema_role == VIEW_ROLE:
        return adapter.view_schema(ordered=True)
    raise httpexceptions.HTTPBadRequest("invalid or missing schema role")


def includeme(config):
    config.include("..predicates")
    config.add_route(
        "cubicweb-jsonschema.schema",
        "*traverse",
        factory=resources.RootResource,
        strict_accept=JSONSCHEMA_MEDIA_TYPE,
    )
    config.scan(__name__)



from base64 import b64encode, b64decode

from cubicweb import _, Binary

from cubicweb_jsonschema.mappers import (
    BytesMapper,
    CompoundMapper,
    StringMapper,
    yams_match,
)


class EMailMapper(StringMapper):
    __select__ = yams_match(
        etype="EmailAddress", rtype="address", role="subject"
    )
    format = "email"


class ThumbnailDataMapper(BytesMapper):
    __select__ = yams_match(
        etype="Thumbnail", rtype="data", role="subject", target_types="Bytes"
    )

    @staticmethod
    def _type(value):
        return Binary(b64decode(value))

    @staticmethod
    def _value(value):
        return b64encode(value.getvalue()).decode("ascii")


class exif_data(CompoundMapper):
    etype = "Photo"
    relations = ("exposure_time", "flash", "maker_note")
    title = _("EXIF data")



"""Tests for "testing" module."""

from functools import partial
import doctest
import unittest

import webtest

from cubicweb_jsonschema import testing


class WebTestReponseOutputCheckerTC(unittest.TestCase):

    def setUp(self):
        self.checker = testing.WebTestReponseOutputChecker()

    def example(self, want):
        """Return a mock of doctest.Example suitable for output_difference()
        usage.
        """
        return type("Example", (object,), {"want": want})()

    def test_ok(self):
        got = want = "\n".join(
            [
                "Response: 200 OK",
                "Content-Type: application/json",
                '{"id": "1", "login": "alice"}',
            ]
        )
        self.assertTrue(self.checker.check_output(want, got, 0))
        self.assertMultiLineEqual(
            self.checker.output_difference(self.example(want), got, 0), ""
        )

    def test_ok_json_body_prettyformat(self):
        got = "\n".join(
            [
                "Response: 200 OK",
                "Content-Type: application/json",
                '{"id": "1", "login": "alice"}',
            ]
        )
        want = "\n".join(
            [
                "Response: 200 OK",
                "Content-Type: application/json",
                "{",
                '  "id": "1",',
                '  "login": "alice"',
                "}",
            ]
        )
        self.assertTrue(self.checker.check_output(want, got, 0))
        self.assertMultiLineEqual(
            self.checker.output_difference(self.example(want), got, 0), ""
        )

    def test_json_ellipsis_object(self):
        got = "\n".join(
            [
                "Response: 200 OK",
                "Content-Type: application/json",
                '{"id": "1", "login": "alice"}',
            ]
        )
        want = "\n".join(
            [
                "Response: 200 OK",
                "Content-Type: application/json",
                "{" '  "id": "...",' '  "login": "alice"' "}",
            ]
        )
        self.assertFalse(self.checker.check_output(want, got, 0))
        self.assertTrue(self.checker.check_output(want, got, doctest.ELLIPSIS))

    def test_json_ellipsis_list(self):
        got = "\n".join(
            [
                "Response: 200 OK",
                "Content-Type: application/json",
                '[{"id": "1", "login": "alice"}]',
            ]
        )
        want = "\n".join(
            [
                "Response: 200 OK",
                "Content-Type: application/json",
                "[" '  {"id": "...", "login": "alice"}' "]",
            ]
        )
        self.assertFalse(self.checker.check_output(want, got, 0))
        self.assertTrue(self.checker.check_output(want, got, doctest.ELLIPSIS))

    def test_bad_json(self):
        got = "\n".join(
            [
                "Response: 200 OK",
                "Content-Type: application/json",
                "{I am invalid JSON!",
            ]
        )
        want = "\n".join(
            [
                "Response: 200 OK",
                "Content-Type: application/json",
                '{"id": "1"}',
            ]
        )
        self.assertFalse(self.checker.check_output(want, got, 0))
        expected = "\n".join(
            [
                "Actual body is invalid JSON: {I am invalid JSON!",
            ]
        )
        self.assertMultiLineEqual(
            self.checker.output_difference(self.example(want), got, 0), expected
        )

    def test_status_differ(self):
        got = "\n".join(
            [
                "Response: 200 OK",
                "Content-Type: application/json",
                '{"id": "1", "login": "alice"}',
            ]
        )
        want = "\n".join(
            [
                "Response: 404 Not Found",
                "Content-Type: application/json",
                "{",
                '  "id": "1",',
                '  "login": "alice"',
                "}",
            ]
        )
        self.assertFalse(self.checker.check_output(want, got, 0))
        expected = "\n".join(
            [
                "Expected status: Response: 404 Not Found",
                "Actual status: Response: 200 OK",
            ]
        )
        self.assertMultiLineEqual(
            self.checker.output_difference(self.example(want), got, 0), expected
        )

    def test_headers_differ(self):
        got = "\n".join(
            [
                "Response: 200 OK",
                "Content-Type: application/octet-stream",
                '{"x": 1}',
            ]
        )
        want = "\n".join(
            [
                "Response: 200 OK",
                "Content-Type: application/json",
                '{"x": 1}',
            ]
        )
        self.assertFalse(self.checker.check_output(want, got, 0))
        expected = "\n".join(
            [
                "Expected headers: Content-Type: application/json",
                "Actual headers: Content-Type: application/octet-stream",
            ]
        )
        self.assertMultiLineEqual(
            self.checker.output_difference(self.example(want), got, 0), expected
        )

    def test_headers_missing_from_got(self):
        got = "\n".join(
            [
                "Response: 200 OK",
                "Content-Type: application/json",
                '{"x": 1}',
            ]
        )
        want = "\n".join(
            [
                "Response: 200 OK",
                "Content-Type: application/json",
                'Link: </schema> rel="describedby"',
                '{"x": 1}',
            ]
        )
        self.assertFalse(self.checker.check_output(want, got, 0))

    def test_headers_extra_in_got(self):
        got = "\n".join(
            [
                "Response: 200 OK",
                "Content-Type: application/json",
                "Something: we do not care about",
                '{"x": 1}',
            ]
        )
        want = "\n".join(
            [
                "Response: 200 OK",
                "Content-Type: application/json",
                '{"x": 1}',
            ]
        )
        self.assertTrue(self.checker.check_output(want, got, 0))

    def test_headers_ellipsis(self):
        got = "\n".join(
            [
                "Response: 200 OK",
                "Location: http://www.example.com/123",
                "Content-Type: application/json",
                '{"x": 1}',
            ]
        )
        want = "\n".join(
            [
                "Response: 200 OK",
                "Content-Type: application/json",
                "Location: http://www.example.com/...",
                '{"x": 1}',
            ]
        )
        self.assertFalse(self.checker.check_output(want, got, 0))
        self.assertTrue(self.checker.check_output(want, got, doctest.ELLIPSIS))
        want = "\n".join(
            [
                "Response: 200 OK",
                "Content-Type: application/json",
                "Location: http://www.example.org/...",
                '{"x": 1}',
            ]
        )
        self.assertFalse(self.checker.check_output(want, got, doctest.ELLIPSIS))

    def test_headers_ellipsis_body_diff(self):
        got = "\n".join(
            [
                "Response: 200 OK",
                "Location: http://www.example.com/123",
                "Content-Type: application/json",
                '{"id": "1"}',
            ]
        )
        want = "\n".join(
            [
                "Response: 200 OK",
                "Content-Type: application/json",
                "Location: http://www.example.com/...",
                '{"id": "2"}',
            ]
        )
        self.assertFalse(self.checker.check_output(want, got, doctest.ELLIPSIS))
        difference = self.checker.output_difference(self.example(want), got, 0)
        # Do not check full content until output_difference() handles ELLIPSIS
        # flag.
        self.assertIn("JSON bodies differ", difference)

    def test_json_body_differ(self):
        got = "\n".join(
            [
                "Response: 200 OK",
                "Content-Type: application/json",
                '{"id": "1"}',
            ]
        )
        want = "\n".join(
            [
                "Response: 200 OK",
                "Content-Type: application/json",
                '{"id": "2"}',
            ]
        )
        self.assertFalse(self.checker.check_output(want, got, 0))
        self.assertFalse(self.checker.check_output(want, got, doctest.ELLIPSIS))
        expected = "\n".join(
            [
                "JSON bodies differ",
                "expected",
                "{",
                '  "id": "2"',
                "}",
                "actual",
                "{",
                '  "id": "1"',
                "}",
                "diff:",
                '{"id": ["2", "1"]}',
            ]
        )
        difference = self.checker.output_difference(self.example(want), got, 0)
        self.assertMultiLineEqual(difference, expected)

    def test_no_content(self):
        got = "Response: 204 No Content<BLANKLINE><BLANKLINE>"
        want = "Response: 204 No Content<BLANKLINE>"
        self.assertTrue(self.checker.check_output(want, got, 0))
        got = "Response: 204 No Content\nFoo: bar\n"
        want = 'Response: 204 No Content\nFoo: bar\n{"x": 1}'
        self.assertFalse(self.checker.check_output(want, got, 0))
        got = 'Response: 204 No Content\nFoo: bar\n{"x": 1}'
        want = "Response: 204 No Content\nFoo: bar\n<BLANKLINE>"
        self.assertFalse(self.checker.check_output(want, got, 0))

    def test_parse_error(self):
        got = "\n".join(
            [
                "Response 42",
            ]
        )
        want = "\n".join(
            [
                "Response: OK",
                "Content-Type: something",
                "bobybody",
            ]
        )
        self.assertFalse(self.checker.check_output(want, got, 0))
        expected = "\n".join(
            [
                "In expected: not a webtest.TestResponse string",
                want,
                "In actual: not a webtest.TestResponse string",
                got,
            ]
        )
        difference = self.checker.output_difference(self.example(want), got, 0)
        self.assertMultiLineEqual(difference, expected)


def unittest_integration_test():
    """Return a DocFileSuite instance using "doctest-example.txt" file with
    a setUp coming from a unittest TestCase.
    """

    class WebTestBaseCase(unittest.TestCase):
        """Base TestCase with a custom setup initializing a webtest
        application.
        """

        def setUp(self):
            def application(environ, start_response):
                body = b'{"a": 1, "b": [1, 2]}'
                headers = [
                    ("Content-Type", "application/json; charset=utf8"),
                    ("Content-Length", str(len(body))),
                ]
                start_response("200 OK", headers)
                return [body]

            self.webapp = webtest.TestApp(application)

    setUp = partial(
        testing.setUp, cls=WebTestBaseCase, globs_from_testcase=["webapp"]
    )

    return doctest.DocFileSuite(
        "doctest-example.txt",
        setUp=setUp,
        checker=testing.WebTestReponseOutputChecker(),
    )


def load_tests(loader, tests, ignore):
    """unittest discovery callback"""
    tests.addTests(
        unittest_integration_test(),
    )
    tests.addTests(
        doctest.DocTestSuite("cubicweb_jsonschema.testing"),
    )
    return tests


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""cubicweb-jsonschema entities tests"""

from base64 import b64encode
from copy import deepcopy

from cubicweb import ValidationError
from cubicweb.devtools.testlib import CubicWebTC

from cubicweb_jsonschema.entities.ijsonschema import (
    jsonschema_adapter,
)


class MiscTC(CubicWebTC):
    """Tests for functions and mixins in entities module."""

    def test_document_tree_nested(self):
        with self.admin_access.cnx() as cnx:
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.entity", cnx, etype="CWUser"
            )
            document = mapper.jsl_document("creation")
            j_schema = document.get_schema()
        self.assertIn(
            "thumbnail", j_schema["definitions"]["Photo"]["properties"]
        )
        self.assertIn("Thumbnail", j_schema["definitions"])


class IJSONSchemaETypeAdapterTC(CubicWebTC):

    maxDiff = None

    def test_adapter_from_etype_uicfg(self):
        with self.admin_access.cnx() as cnx:
            adapter = jsonschema_adapter(cnx, etype="CWUser")
            self.assertIsNotNone(adapter)
            j_schema = adapter.creation_schema()
        user_props = j_schema["properties"]
        self.assertIn("firstname", user_props)
        self.assertNotIn("lastname", user_props)  # per uicfg

    def test_adapter_from_etype_view_schema(self):
        with self.admin_access.cnx() as cnx:
            adapter = jsonschema_adapter(cnx, etype="CWUser")
            j_schema = adapter.view_schema()
        self.assertNotIn("EmailAlias", j_schema["definitions"])  # per uicfg
        self.assertNotIn("required", j_schema)
        user_props = j_schema["properties"]
        self.assertIn("use_email", user_props)
        self.assertNotIn("upassword", user_props)
        emailaddr_def = j_schema["definitions"]["EmailAddress"]
        self.assertNotIn("use_email", emailaddr_def["properties"])

    def test_relation_in_attributes_section(self):
        with self.admin_access.cnx() as cnx:
            adapter = jsonschema_adapter(cnx, etype="CWUser")
            creation_schema = adapter.creation_schema()
            view_schema = adapter.view_schema()
            groups = [
                (group.eid, group.dc_title())
                for group in cnx.find("CWGroup").entities()
                if group.name != "owners"
            ]
        self.assertNotIn("CWGroup", creation_schema["definitions"])
        user_creation_schema = creation_schema["properties"]
        expected_items = {
            "type": "string",
            "oneOf": [
                {"enum": [str(eid)], "title": title} for eid, title in groups
            ],
        }
        self.assertEqual(
            user_creation_schema["in_group"]["items"], expected_items
        )
        expected = {
            "type": "array",
            "items": expected_items,
            "description": "groups grant permissions to the user",
            "title": "in_group",
        }
        self.assertEqual(view_schema["properties"]["in_group"], expected)

    def test_target_of_relation_in_inlined_autoform_section(self):
        with self.admin_access.cnx() as cnx:
            adapter = jsonschema_adapter(cnx, etype="CWUser")
            j_schema = adapter.creation_schema()
        self.assertIn("EmailAddress", j_schema["definitions"])

    def test_target_of_relation_has_a_bytes_attribute(self):
        with self.admin_access.cnx() as cnx:
            adapter = jsonschema_adapter(cnx, etype="CWUser")
            j_schema = adapter.creation_schema()
        self.assertIn("Photo", j_schema["definitions"])
        photo_schema = j_schema["definitions"]["Photo"]
        self.assertEqual(photo_schema["properties"]["data"]["type"], "string")

    def test_target_of_relation_of_relation_registered(self):
        with self.admin_access.cnx() as cnx:
            adapter = jsonschema_adapter(cnx, etype="CWUser")
            j_schema = adapter.creation_schema()
        self.assertIn("Thumbnail", j_schema["definitions"])

    def test_create_entity(self):
        with self.admin_access.cnx() as cnx:
            group = cnx.find("CWGroup", name="users").one()
            adapter = jsonschema_adapter(cnx, etype="CWUser")
            instance = {
                "login": "bob",
                "upassword": "123",
                "firstname": "Bob",
                "in_group": [str(group.eid)],
            }
            user = adapter.create_entity(instance)
            cnx.commit()
            for attrname, value in instance.items():
                self.assertEqual(getattr(user, attrname), value)

    def test_create_entity_default_required(self):
        """Not supplying a required property which schema has a "default" flag
        is allowed.
        """
        with self.admin_access.cnx() as cnx:
            adapter = jsonschema_adapter(cnx, etype="Photo")
            j_schema = adapter.creation_schema()
            self.assertIn("media_type", j_schema["required"])
            instance = {
                "data": "plop",
            }
            entity = adapter.create_entity(instance)
            cnx.commit()
            self.assertEqual(entity.media_type, "png")

    def test_create_entity_inlined(self):
        with self.admin_access.cnx() as cnx:
            adapter = jsonschema_adapter(cnx, etype="Photo")
            instance = {
                "data": "plop",
                "media_type": "png",
                "thumbnail": [
                    {
                        "data": b64encode(b"plip").decode("ascii"),
                    },
                ],
            }
            expected = deepcopy(instance)
            expected["exif_data"] = {
                "flash": False,
            }
            entity = adapter.create_entity(instance.copy())
            entity.cw_clear_all_caches()
            self.assertEqual(len(entity.thumbnail), 1)
            self.assertEqual(entity.thumbnail[0].data.getvalue(), b"plip")
            serialized = entity.cw_adapt_to("IJSONSchema").serialize()
            self.assertEqual(serialized, expected)

    def test_schema_role_required(self):
        """Depending on schema role, required field should not be set."""
        with self.admin_access.cnx() as cnx:
            adapter = jsonschema_adapter(cnx, etype="CWUser")
            creation_schema = adapter.creation_schema()
            view_schema = adapter.view_schema()
        # 'required' only present in 'creation' schema.
        self.assertNotIn("required", view_schema)
        self.assertIn("properties", view_schema)
        self.assertCountEqual(
            creation_schema["required"], ["login", "upassword", "in_group"]
        )


class IJSONSchemaRelationTargetETypeAdapterTC(CubicWebTC):

    def test_creation_set_relation(self):
        with self.admin_access.cnx() as cnx:
            adapter = jsonschema_adapter(
                cnx, etype="CWUser", rtype="in_group", role="subject"
            )
            group = cnx.find("CWGroup", name="users").one()
            user = adapter.create_entity(
                {"login": "bob", "upassword": "bob"}, group
            )
            cnx.commit()
            self.assertEqual([x.eid for x in user.in_group], [group.eid])
            with self.assertRaises(ValidationError) as cm:
                instance = {
                    "login": "bob",
                    "upassword": "bob",
                    "in_group": [group.eid],
                }
                adapter.create_entity(instance, group)
                cnx.commit()
            self.assertIn("unexpected properties: in_group", str(cm.exception))


class IJSONSchemaEntityAdapterTC(CubicWebTC):

    def test_adapter_from_entity(self):
        with self.admin_access.cnx() as cnx:
            user = cnx.find("CWUser", login="admin").one()
            adapter = user.cw_adapt_to("IJSONSchema")
            self.assertIsNotNone(adapter)
            j_schema = adapter.edition_schema()
        self.assertNotIn("lastname", j_schema["properties"])

    def test_edition_schema_role(self):
        with self.admin_access.cnx() as cnx:
            user = cnx.find("CWUser", login="admin").one()
            adapter = user.cw_adapt_to("IJSONSchema")
            j_schema = adapter.edition_schema()
            user_props = j_schema["properties"]
            self.assertIn("upassword", user_props)
            # in "inlined" jsonschema_section
            self.assertIn("use_email", user_props)
            # per jsonschema_section
            self.assertNotIn("lastname", user_props)

    def test_relations(self):
        with self.admin_access.cnx() as cnx:
            entity = cnx.create_entity("Book", title="bouquin")
            cnx.commit()
            adapter = entity.cw_adapt_to("IJSONSchema")
            relations = list(adapter.relations())
            expected = [
                ("author", "subject", set(["Author"])),
                ("publications", "object", set(["Author"])),
                ("topics", "subject", set(["Topic"])),
            ]
            self.assertCountEqual(relations, expected)

    def test_entity_create(self):
        with self.admin_access.cnx() as cnx:
            users = cnx.find("CWGroup", name="users").one()
            adapter = jsonschema_adapter(cnx, etype="CWUser")
            instance = {
                "login": "bob",
                "upassword": "sponge",
                "in_group": [str(users.eid)],
                "use_email": [{"address": "bob@sponge.com"}],
            }
            entity = adapter.create_entity(instance)
            self.assertEqual(entity.login, "bob")
            self.assertEqual(entity.upassword, b"sponge")
            self.assertEqual(len(entity.in_group), 1)
            self.assertEqual(entity.in_group[0].name, "users")
            self.assertEqual(len(entity.use_email), 1)
            self.assertEqual(entity.use_email[0].address, "bob@sponge.com")

    def test_entity_update(self):
        with self.admin_access.cnx() as cnx:
            entity = self.create_user(cnx, "bob", password="sponge")
            cnx.commit()
            users = cnx.find("CWGroup", name="users").one()
            guests = cnx.find("CWGroup", name="guests").one()
            adapter = jsonschema_adapter(cnx, entity=entity)

            instance = {
                "login": "bobby",
                "in_group": [str(users.eid)],
                "use_email": [{"address": "bob@sponge.com"}],
            }
            adapter.edit_entity(instance)
            entity.cw_clear_all_caches()
            self.assertEqual(entity.login, "bobby")
            # ensure password have not been reseted
            with cnx.security_enabled(read=False):
                self.assertTrue(entity.upassword)
            self.assertEqual(len(entity.in_group), 1)
            self.assertEqual(entity.in_group[0].name, "users")
            self.assertEqual(len(entity.use_email), 1)
            self.assertEqual(entity.use_email[0].address, "bob@sponge.com")

            instance = {
                "login": "bobby",
                "in_group": [str(users.eid), str(guests.eid)],
                "use_email": [{"address": "bobby@sponge.com"}],
            }
            adapter.edit_entity(instance)
            entity.cw_clear_all_caches()
            self.assertEqual(entity.login, "bobby")
            self.assertEqual(len(entity.in_group), 2)
            self.assertCountEqual(
                [group.name for group in entity.in_group], ["users", "guests"]
            )
            self.assertEqual(len(entity.use_email), 1)
            self.assertEqual(entity.use_email[0].address, "bobby@sponge.com")

            instance = {
                "login": "bobby",
                "in_group": [str(users.eid)],
                "use_email": [
                    {"address": "bobby@sponge.com"},
                    {"address": "bob.sponge@sponge.com"},
                ],
            }
            adapter.edit_entity(instance)
            entity.cw_clear_all_caches()
            self.assertEqual(len(entity.in_group), 1)
            self.assertEqual(entity.in_group[0].name, "users")
            self.assertEqual(len(entity.use_email), 2)
            self.assertCountEqual(
                [addr.address for addr in entity.use_email],
                ["bobby@sponge.com", "bob.sponge@sponge.com"],
            )

            instance = {
                "login": "bobby",
                "in_group": [str(users.eid)],
                "use_email": [{"address": "bob.sponge@sponge.com"}],
            }
            adapter.edit_entity(instance)
            entity.cw_clear_all_caches()
            self.assertEqual(len(entity.use_email), 1)
            self.assertEqual(
                entity.use_email[0].address, "bob.sponge@sponge.com"
            )

            entity.cw_set(use_email=cnx.create_entity("EmailAlias"))
            instance = {
                "login": "bobby",
                "in_group": [str(users.eid)],
                "use_email": [],
            }
            adapter.edit_entity(instance)
            entity.cw_clear_all_caches()
            self.assertEqual(len(entity.use_email), 1)
            self.assertEqual(entity.use_email[0].cw_etype, "EmailAlias")

    def test_serialize(self):
        self.maxDiff = None
        with self.admin_access.cnx() as cnx:
            entity = self.create_user(
                cnx, "bob", password="sponge", firstname="Bob"
            )
            email = cnx.create_entity(
                "EmailAddress",
                address="bob@sponge.com",
                reverse_use_email=entity,
            )
            cnx.commit()
            entity.cw_clear_all_caches()
            email.cw_clear_all_caches()

            groups = entity.related("in_group", "subject").entities()
            group_ids = [str(group.eid) for group in groups]
            expected = {
                "firstname": "Bob",
                "login": "bob",
                "in_group": group_ids,
                "use_email": [{"address": "bob@sponge.com"}],
            }
            data = entity.cw_adapt_to("IJSONSchema").serialize()
            self.assertEqual(data, expected)


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2017-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""cubicweb-jsonschema's site configuration."""


def add_mappers_to_cube_appobject_path() -> None:
    """Add "mappers" to `cube_appobject_path` class attribute of CubicWeb
    configuration classes.
    """
    import importlib
    import logging

    LOGGER = logging.getLogger("cubicweb-jsonschema")
    LOGGER.info('adding "mappers" to cube appobjects path')

    # See cubicweb.cwconfig.CubicWebConfiguration.load_available_configs()
    # where all known configuration modules are listed.
    # TODO: We should have a better way to discovered registered
    # configuration classes.
    CONFS = (
        ("cubicweb.pyramid.config", "AllInOneConfiguration"),
        ("cubicweb.cw_config", "CubicWebConfiguration"),
        ("cubicweb.devtools.apptest_config", "ApptestConfiguration"),
        ("cubicweb_web.devtools.testlib", "WebApptestConfiguration"),
        (
            "cubicweb_web.devtools.testlib",
            "WebPostgresApptestConfiguration",
        ),
        ("cubicweb_web.webconfig", "WebAllInOneConfiguration"),
    )

    for modname, configclsname in CONFS:
        try:
            mod = importlib.import_module(modname)
        except ImportError:
            LOGGER.warning(
                'failed to import %s, "mappers" appobjects will not '
                "loaded for this configuration",
                modname,
            )
            continue
        configcls = getattr(mod, configclsname)
        configcls.cube_appobject_path.add("mappers")


add_mappers_to_cube_appobject_path()
del add_mappers_to_cube_appobject_path



# pylint: disable=W0622
"""cubicweb-jsonschema application packaging information"""


cubename = "jsonschema"
modname = "cubicweb_" + cubename
distname = "cubicweb-" + cubename

numversion = (0, 2, 14)
version = ".".join(str(num) for num in numversion)

license = "LGPL"
author = "LOGILAB S.A. (Paris, FRANCE)"
author_email = "contact@logilab.fr"
description = "JSON Schema for CubicWeb"
web = "https://forge.extranet.logilab.fr/cubicweb/cubes/" + cubename

__depends__ = {
    "cubicweb": ">= 4.5.2, < 5.0.0",
    "cubicweb-web": ">= 1.3.1, < 2.0.0",
    "iso8601": None,
    "jsl": None,
    "pyramid": ">= 1.10.8",
}
__recommends__ = {}

classifiers = [
    "Environment :: Web Environment",
    "Framework :: CubicWeb",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: JavaScript",
]



# copyright 2016-2022 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

import base64
from datetime import datetime
import unittest

import jsl
import jsonschema
from pytz import UTC

from cubicweb import (
    Binary,
    Unauthorized,
    _,
)
from cubicweb.devtools.testlib import CubicWebTC

from cubicweb_jsonschema import CREATION_ROLE, EDITION_ROLE, VIEW_ROLE
from cubicweb_jsonschema.mappers import CompoundMapper, yams_match


class CollectionMapperTC(CubicWebTC):

    def test_jsl_field_view(self):
        with self.admin_access.cnx() as cnx:
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.collection", cnx, etype="Book"
            )
            field = mapper.jsl_field(VIEW_ROLE)
        j_schema = field.get_schema()
        expected = {
            "type": "array",
            "title": "Book_plural",
            "items": {
                "properties": {
                    "type": {"type": "string"},
                    "id": {"type": "string"},
                    "title": {"type": "string"},
                },
                "type": "object",
            },
        }
        self.assertEqual(j_schema, expected)

    def test_jsl_field_creation(self):
        with self.admin_access.cnx() as cnx:
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.collection", cnx, etype="Book"
            )
            field = mapper.jsl_field(CREATION_ROLE)
        j_schema = field.get_schema()
        expected = {
            "type": "array",
            "title": "Book_plural",
            "items": {"type": "string"},
        }
        self.assertEqual(j_schema, expected)


class RelationMapperTC(CubicWebTC):

    def test_target_types(self):
        with self.admin_access.client_cnx() as cnx:
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.relation",
                cnx,
                etype="CWEType",
                rtype="add_permission",
                role="subject",
                target_types={"CWGroup", "RQLExpression"},
            )
            self.assertCountEqual(
                mapper.target_types, ["CWGroup", "RQLExpression"]
            )

            mapper = cnx.vreg["mappers"].select(
                "jsonschema.relation",
                cnx,
                etype="CWEType",
                rtype="add_permission",
                role="subject",
                target_types={"CWGroup"},
            )
            self.assertEqual(mapper.target_types, ["CWGroup"])

    def test_multiple_target_types_inlined(self):
        with self.admin_access.client_cnx() as cnx:
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.relation",
                cnx,
                etype="CWUser",
                rtype="use_email",
                role="subject",
                target_types={"EmailAddress", "EmailAlias"},
            )
            field = mapper.jsl_field(CREATION_ROLE)
            self.assertIsInstance(field, jsl.fields.ArrayField)
            self.assertIsInstance(field.items, jsl.fields.OneOfField)
            self.assertEqual(len(field.items.fields), 2)

    def test_attribute_have_no_target_types(self):
        with self.admin_access.client_cnx() as cnx:
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.relation",
                cnx,
                etype="CWUser",
                rtype="login",
                role="subject",
                target_types={"String"},
            )
            self.assertCountEqual(mapper.target_types, [])

    def test_password_field_not_required_on_update(self):
        with self.admin_access.client_cnx() as cnx:
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.relation",
                cnx,
                etype="CWUser",
                rtype="upassword",
                role="subject",
                target_types={"Password"},
            )
            self.assertTrue(mapper.jsl_field(CREATION_ROLE).required)
            self.assertFalse(mapper.jsl_field(EDITION_ROLE).required)

    def test_inlined_relation_select_on_entity(self):
        with self.admin_access.client_cnx() as cnx:
            user = cnx.find("CWUser", login="admin").one()
            mapper = cnx.vreg["mappers"].select_or_none(
                "jsonschema.relation",
                cnx,
                entity=user,
                rtype="use_email",
                role="subject",
                target_types={"EmailAddress"},
            )
            self.assertIsNotNone(mapper)

    def test_vocabulary(self):
        with self.admin_access.cnx() as cnx:
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.relation",
                cnx,
                etype="Photo",
                rtype="media_type",
                role="subject",
                target_types={"String"},
            )
            field = mapper.jsl_field(CREATION_ROLE)
            expected = {
                "oneOf": [
                    {"enum": ["jpeg"], "title": "jpeg", "type": "string"},
                    {"enum": ["png"], "title": "png", "type": "string"},
                ],
                "default": "png",
                "title": "media_type",
            }
            self.assertEqual(field.get_schema(), expected)
            adapter = self.vreg["adapters"].select(
                "IJSONSchema", cnx, etype="Photo"
            )
            # Make sure 'required' constraint is present in the main schema.
            schema = adapter.creation_schema()
            self.assertIn("media_type", schema["required"])

    def test_required_generic_relation(self):
        with self.admin_access.cnx() as cnx:
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.entity", cnx, etype="CWUser"
            )
            schema = mapper.jsl_document(CREATION_ROLE).get_schema()
            self.assertIn("in_group", schema["required"])


class InlinedRelationMapperTC(CubicWebTC):

    def test_schema(self):
        with self.admin_access.client_cnx() as cnx:
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.relation",
                cnx,
                etype="CWUser",
                rtype="use_email",
                role="subject",
                target_types={"EmailAddress"},
            )
            field = mapper.jsl_field(CREATION_ROLE)
            schema = field.get_schema()
            expected = {
                "title": "use_email",
                "type": "array",
                "items": {
                    "$ref": "#/definitions/EmailAddress",
                },
                "description": "user's email account",
                "definitions": {
                    "EmailAddress": {
                        "title": "EmailAddress",
                        "type": "object",
                        "additionalProperties": False,
                        "properties": {
                            "address": {
                                "format": "email",
                                "maxLength": 128,
                                "title": "address",
                                "type": "string",
                            },
                            "alias": {
                                "maxLength": 56,
                                "title": "alias",
                                "type": "string",
                            },
                        },
                        "required": ["address"],
                    },
                },
            }
        self.maxDiff = None
        self.assertEqual(schema, expected)


class AttributeMappersTC(CubicWebTC):

    def test_bytes_serialize(self):
        with self.admin_access.cnx() as cnx:
            entity = cnx.create_entity("Photo", data=Binary(b"plop"))
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.relation",
                cnx,
                etype="Photo",
                rtype="data",
                role="subject",
                target_types={"Bytes"},
            )
            self.assertEqual(mapper.__class__.__name__, "BytesMapper")
            self.assertEqual(mapper.serialize(entity), "plop")

    def test_bytes_value(self):
        with self.admin_access.cnx() as cnx:
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.relation",
                cnx,
                etype="Photo",
                rtype="data",
                role="subject",
                target_types={"Bytes"},
            )
            self.assertEqual(mapper.__class__.__name__, "BytesMapper")
            values = mapper.values(None, {"data": "plop"})
            self.assertEqual(len(values), 1)
            bin_value = values["data"]
            self.assertIsInstance(bin_value, Binary)
            self.assertEqual(bin_value.getvalue(), b"plop")

    def test_date_value(self):
        with self.admin_access.cnx() as cnx:
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.relation",
                cnx,
                etype="CWUser",
                rtype="last_login_time",
                role="subject",
                target_types={"TZDatetime"},
            )
            instance = {"last_login_time": "2007-01-25T12:00:00Z"}
            values = mapper.values(None, instance)
            expected = {
                "last_login_time": datetime(2007, 1, 25, 12, 0, tzinfo=UTC),
            }
            self.assertEqual(values, expected)

    def test_password_values(self):
        with self.admin_access.cnx() as cnx:
            users = cnx.find("CWGroup", name="users").one()
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.relation",
                cnx,
                etype="CWUser",
                rtype="upassword",
                role="subject",
                target_types={"Password"},
            )
            instance = {
                "upassword": "bob",
            }
            values = mapper.values(None, instance)
            self.assertEqual(values, {"upassword": b"bob"})

            # Make sure we can create an entity with these "values".
            user = cnx.create_entity(
                "CWUser", login="bob", in_group=users, **values
            )
            cnx.commit()
            self.assertEqual(user.upassword, b"bob")

            # Now check update.
            instance = {
                "upassword": "bobby",
            }
            values = mapper.values(user, instance)
            self.assertEqual(values, {"upassword": b"bobby"})
            # We cannot apparently read the password after entity creation, so
            # just make sure update does not raise a validation error.
            try:
                user.cw_set(**values)
                cnx.commit()
            except Exception as exc:
                self.fail(str(exc))

    def test_select_custom_mapper_over_default_one(self):
        with self.admin_access.cnx() as cnx:
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.relation",
                cnx,
                etype="Thumbnail",
                rtype="data",
                role="subject",
                target_types={"Bytes"},
            )
            self.assertEqual(mapper.__class__.__name__, "ThumbnailDataMapper")
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.relation",
                cnx,
                etype="Photo",
                rtype="data",
                role="subject",
                target_types={"Bytes"},
            )
            self.assertEqual(mapper.__class__.__name__, "BytesMapper")

    def test_custom_mapper_for_bytes(self):
        with self.admin_access.cnx() as cnx:
            thumbnail = cnx.create_entity("Thumbnail", data=Binary(b"plip"))
            photo = cnx.create_entity(
                "Photo", data=Binary(b"plop"), thumbnail=thumbnail
            )
            cnx.commit()
            adapter = photo.cw_adapt_to("IJSONSchema")
            instance = adapter.serialize()
            expected = {
                "data": "plop",
                "exif_data": {
                    "flash": False,
                },
                "media_type": "png",
                "thumbnail": [
                    {"data": base64.b64encode(b"plip").decode("ascii")},
                ],
            }
            self.assertEqual(instance, expected)

    def test_custom_format(self):
        with self.admin_access.cnx() as cnx:
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.relation",
                cnx,
                etype="EmailAddress",
                rtype="address",
                role="subject",
                target_types={"String"},
            )
            schema = mapper.jsl_field(VIEW_ROLE).get_schema()
            expected = {
                "type": "string",
                "title": "address",
                "format": "email",
            }
            self.assertEqual(schema, expected)

    def test_unchanged_value_not_updated(self):
        with self.admin_access.cnx() as cnx:
            self.create_user(cnx, "bob")
            cnx.commit()
        with self.new_access("bob").cnx() as cnx:
            entity = cnx.create_entity("Topic", name="top")
            cnx.commit()
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.relation",
                cnx,
                entity=entity,
                rtype="name",
                role="subject",
                target_types={"String"},
            )
            # By permission, we can change the attribute.
            values = mapper.values(entity, {"name": "tip"})
            self.assertEqual(values, {"name": "tip"})
            entity.cw_set(**values)
            cnx.commit()
            cnx.create_entity("Book", title="ti", topics=entity)
            cnx.commit()
            # Value change, we get something returned.
            values = mapper.values(entity, {"name": "top"})
            self.assertEqual(values, {"name": "top"})
            # By permission, we can't change the attribute.
            with self.assertRaises(Unauthorized):
                entity.cw_set(**values)
                cnx.commit()
            cnx.rollback()
            # But passing the actual value of attribute should not trigger
            # entity update.
            values = mapper.values(entity, {"name": "tip"})
            self.assertEqual(values, {})


class CompoundMapperTC(CubicWebTC):

    def mapper_by_name(self, name):
        for obj in self.vreg["mappers"]["jsonschema.object"]:
            if obj.name == name:
                return obj
        raise ValueError(name)

    def test_no_relations_attribute(self):
        class bad_compound(CompoundMapper):
            etype = "Photo"

        expected_regexp = 'is missing a "relations" class attribute'
        with self.assertRaisesRegex(ValueError, expected_regexp):
            bad_compound.__registered__(self.vreg["mappers"])

    def test_no_duplicate_relation_mapping(self):
        """A CompoundMapper with the same relation 'flash' is defined in
        test/data/mappers.py, thus this one is a duplicate and should fail.
        """

        class bad_compound(CompoundMapper):
            etype = "Photo"
            relations = ("flash",)

        expected_regexp = 'duplicate relation mapping for "Photo": flash'
        with self.assertRaisesRegex(ValueError, expected_regexp):
            bad_compound.__registered__(self.vreg["mappers"])

    def test_no_duplicate_subject_relation_mapping(self):
        """A CompoundMapper with the same relation 'flash' is defined in
        test/data/mappers.py, thus this one is a duplicate and should fail.
        """

        class bad_compound(CompoundMapper):
            etype = "Photo"
            relations = (("flash", "subject"),)

        expected_regexp = 'duplicate relation mapping for "Photo": flash-subject'  # noqa: E501
        with self.assertRaisesRegex(ValueError, expected_regexp):
            bad_compound.__registered__(self.vreg["mappers"])

    def test_no_duplicate_object_relation_mapping(self):
        class good_compound(CompoundMapper):
            etype = "Book"
            relations = (
                "author",
                ("author", "object"),
            )

        try:
            good_compound.__registered__(self.vreg["mappers"])
        except ValueError as e:
            self.fail("unexpected {}".format(e))

    def test_mapped_attributes_hidden(self):
        """Make sure attributes mapped to CompoundMapper are in "hidden" uicfg
        section.
        """
        with self.admin_access.cnx() as cnx:
            entity = cnx.vreg["etypes"].etype_class("Photo")(cnx)
            rsection = cnx.vreg["uicfg"].select(
                "jsonschema", cnx, entity=entity
            )
            for action in ("read", "add"):
                with self.subTest(action=action):
                    hidden = list(
                        rsection.relations_by_section(entity, "hidden", action)
                    )
                    self.assertIn(
                        ("exposure_time", "subject", set(["Float"])), hidden
                    )
                    self.assertIn(
                        ("flash", "subject", set(["Boolean"])), hidden
                    )

    def test_schema(self):
        with self.admin_access.cnx() as cnx:
            mapper = self.mapper_by_name("exif_data")(cnx)
            doc = mapper.jsl_field(VIEW_ROLE)
            schema = doc.get_schema()
            expected = {
                "$ref": "#/definitions/EXIF data",
                "definitions": {
                    "EXIF data": {
                        "title": "EXIF data",
                        "type": "object",
                        "properties": {
                            "exposure_time": {
                                "title": "exposure_time",
                                "type": "number",
                            },
                            "flash": {
                                "title": "flash",
                                "type": "boolean",
                            },
                            "maker_note": {
                                "title": "maker_note",
                                "type": "string",
                            },
                        },
                        "additionalProperties": False,
                    },
                },
            }
            self.maxDiff = None
            self.assertEqual(schema, expected)
            doc = mapper.jsl_field(CREATION_ROLE)
            schema = doc.get_schema()
            expected["definitions"]["EXIF data"]["required"] = ["flash"]
            expected["definitions"]["EXIF data"]["properties"]["flash"][
                "default"
            ] = False
            self.assertEqual(schema, expected)

    def test_serialize(self):
        with self.admin_access.cnx() as cnx:
            entity = cnx.create_entity(
                "Photo", data=Binary(b"plop"), exposure_time=1.23, flash=False
            )
            mapper = self.mapper_by_name("exif_data")(cnx)
            data = mapper.serialize(entity)
            expected = {
                "exposure_time": 1.23,
                "flash": False,
            }
            self.assertEqual(data, expected)
            entity.cw_set(exposure_time=None)
            del expected["exposure_time"]
            data = mapper.serialize(entity)
            self.assertEqual(data, expected)

    def test_values(self):
        with self.admin_access.cnx() as cnx:
            instance = {
                "data": "plop",
                "exif_data": {
                    "flash": True,
                    "exposure_time": 0.87,
                    "maker_note": "secret thing",
                },
            }
            mapper = self.mapper_by_name("exif_data")(cnx)
            values = mapper.values(None, instance.copy())
            self.assertEqual(values, instance["exif_data"])
            self.assertIsInstance(values["maker_note"], Binary)
            self.assertEqual(values["maker_note"].getvalue(), b"secret thing")
            # Now make sure we can actually create an entity from "values".
            # We need to add 'data' property to "values" as the latter comes
            # from the compound mapper which does not manage the former.
            assert "data" not in values
            values["data"] = Binary(b"plop")
            try:
                cnx.create_entity("Photo", **values)
            except Exception as e:
                self.fail(str(e))

    def test_etype_schema(self):
        with self.admin_access.cnx() as cnx:
            adapter = cnx.vreg["adapters"].select(
                "IJSONSchema", cnx, etype="Photo"
            )
            view_schema = adapter.view_schema()
            self._check_schema(view_schema)
            creation_schema = adapter.creation_schema()
            self._check_schema(creation_schema, True)

    def test_entity_schema(self):
        with self.admin_access.cnx() as cnx:
            entity = cnx.create_entity("Photo", data=Binary(b"plop"))
            adapter = entity.cw_adapt_to("IJSONSchema")
            view_schema = adapter.view_schema()
            self._check_schema(view_schema)
            edition_schema = adapter.edition_schema()
            self._check_schema(edition_schema, True)

    def _check_schema(self, schema, edition_role=False):
        self.assertIn("exif_data", schema["properties"])
        self.maxDiff = None
        expected = {
            "title": "EXIF data",
            "type": "object",
            "properties": {
                "exposure_time": {
                    "title": "exposure_time",
                    "type": "number",
                },
                "flash": {
                    "title": "flash",
                    "type": "boolean",
                },
                "maker_note": {
                    "title": "maker_note",
                    "type": "string",
                },
            },
            "additionalProperties": False,
        }
        if edition_role:
            expected["required"] = ["flash"]
            expected["properties"]["flash"]["default"] = False
        self.assertEqual(schema["definitions"]["EXIF data"], expected)

    def test_entity_serialize(self):
        with self.admin_access.cnx() as cnx:
            entity = cnx.create_entity(
                "Photo", data=Binary(b"plop"), exposure_time=1.23, flash=False
            )
            adapter = entity.cw_adapt_to("IJSONSchema")
            schema = adapter.view_schema()
            self._check_schema(schema)
            instance = adapter.serialize()
            try:
                jsonschema.validate(instance, schema)
            except jsonschema.ValidationError as exc:
                self.fail(str(exc))

    def test_entity_create(self):
        with self.admin_access.cnx() as cnx:
            adapter = cnx.vreg["adapters"].select(
                "IJSONSchema", cnx, etype="Photo"
            )
            self._check_insert(cnx, adapter.create_entity)

    def test_entity_create_no_compound_data(self):
        with self.admin_access.cnx() as cnx:
            adapter = cnx.vreg["adapters"].select(
                "IJSONSchema", cnx, etype="Photo"
            )
            instance = {
                "data": "plip",
                "media_type": "jpeg",
            }
            adapter.create_entity(instance)
            cnx.commit()
            entity = cnx.find("Photo").one()
            self.assertEqual(entity.data.getvalue(), b"plip")
            self.assertEqual(entity.flash, False)  # default value

    def test_entity_update(self):
        with self.admin_access.cnx() as cnx:
            entity = cnx.create_entity("Photo", data=Binary(b"plop"))
            cnx.commit()
            adapter = entity.cw_adapt_to("IJSONSchema")
            self._check_insert(cnx, adapter.edit_entity)

    def _check_insert(self, cnx, insert_function):
        instance = {
            "data": "plip",
            "media_type": "jpeg",
            "exif_data": {
                "flash": True,
                "exposure_time": 0.9,
            },
        }
        insert_function(instance)
        cnx.commit()
        entity = cnx.find("Photo").one()
        self.assertEqual(entity.flash, True)
        self.assertEqual(entity.exposure_time, 0.9)
        self.assertEqual(entity.media_type, "jpeg")

    def test_nonfinal_relation_schema(self):

        class bin_data(CompoundMapper):
            etype = "Photo"
            relations = ("data", "thumbnail")
            title = _("Photo data")

        expected_schema = {
            "$ref": "#/definitions/Photo data",
            "definitions": {
                "Photo data": {
                    "title": "Photo data",
                    "type": "object",
                    "properties": {
                        "data": {
                            "title": "data",
                            "type": "string",
                        },
                        "thumbnail": {
                            "items": {
                                "$ref": "#/definitions/Thumbnail",
                            },
                            "title": "thumbnail",
                            "type": "array",
                        },
                    },
                    "additionalProperties": False,
                },
                "Thumbnail": {
                    "title": "Thumbnail",
                    "type": "object",
                    "properties": {
                        "data": {
                            "title": "data",
                            "type": "string",
                        },
                    },
                    "additionalProperties": False,
                },
            },
        }

        with self.temporary_appobjects(bin_data):
            with self.admin_access.cnx() as cnx:
                mapper = self.mapper_by_name("bin_data")(cnx)
                schema = mapper.jsl_field(VIEW_ROLE).get_schema()
                self.assertEqual(schema, expected_schema)

    def test_nonfinal_relation_values(self):

        class bin_data(CompoundMapper):
            etype = "Photo"
            relations = ("data", "thumbnail")
            title = _("Photo data")

        with self.temporary_appobjects(bin_data):
            with self.admin_access.cnx() as cnx:
                mapper = self.mapper_by_name("bin_data")(cnx)
                instance = {
                    "bin_data": {
                        "data": "plip",
                        "thumbnail": [
                            {
                                "data": "plop",
                            },
                        ],
                    },
                }
                values = mapper.values(None, instance)
                self.assertCountEqual(list(values), ["data", "thumbnail"])
                # A Thumbnail entity should have been created.
                thumbnail = cnx.find("Thumbnail").one()
                self.assertEqual(values["thumbnail"], [thumbnail])

    def test_object_relation(self):

        class my_compound(CompoundMapper):
            name = "My Compound"
            title = "My Title"
            etype = "Author"
            relations = (("author", "object"),)

        expected_schema = {
            "$schema": "http://json-schema.org/draft-06/schema#",
            "type": "object",
            "title": "Author",
            "additionalProperties": False,
            "properties": {
                "name": {"title": "name", "type": "string"},
                "My Compound": {
                    "$ref": "#/definitions/My Title",
                },
            },
            "definitions": {
                "My Title": {
                    "type": "object",
                    "title": "My Title",
                    "additionalProperties": False,
                    "properties": {
                        "reverse_author": {
                            "title": "author_object",
                            "items": {
                                "oneOf": [],
                                "type": "string",
                            },
                            "type": "array",
                        },
                    },
                },
            },
        }

        with self.temporary_appobjects(my_compound):
            with self.admin_access.cnx() as cnx:
                author = cnx.create_entity("Author", name="author")
                cnx.create_entity("Book", author=author, title="title")

                adapter = author.cw_adapt_to("IJSONSchema")
                schema = adapter.view_schema()
                self.assertEqual(expected_schema, schema)


class EntityMapperTC(CubicWebTC):

    def test__object_mappers(self):
        with self.admin_access.cnx() as cnx:
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.entity", cnx, etype="Photo"
            )
            self.assertEqual(
                [f.name for f in mapper._object_mappers()], ["exif_data"]
            )

    def test_etypemapper_relations(self):
        expected_rtypes = [
            "login",
            "firstname",
            "surname",
            "last_login_time",
            "in_group",
            "use_email",
            "picture",
        ]
        with self.admin_access.cnx() as cnx:
            user = cnx.find("CWUser", login="admin").one()
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.entity", cnx, entity=user
            )
            rtypes = [rtype for rtype, _, _ in mapper.relations(VIEW_ROLE)]
            self.assertCountEqual(expected_rtypes, rtypes)
            self.create_user(cnx, "bob")
            cnx.commit()
        with self.new_access("bob").cnx() as cnx:
            user = cnx.find("CWUser", login="admin").one()
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.entity", cnx, entity=user
            )
            rtypes = [rtype for rtype, _, _ in mapper.relations(VIEW_ROLE)]
            self.assertCountEqual(expected_rtypes, rtypes)

    def test_etypemapper_relations_accounts_for_uicfg(self):
        with self.admin_access.cnx() as cnx:
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.entity", cnx, etype="CWUser"
            )
            fields = dict(
                (rtype, targets)
                for rtype, _, targets in mapper.relations(VIEW_ROLE)
            )
            # EmailAlias no in, per jsonschema_section.
            self.assertEqual(fields["use_email"], {"EmailAddress"})

    def test_targetetypemapper_relations_and_schema(self):
        with self.admin_access.cnx() as cnx:
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.entity",
                cnx,
                etype="Thumbnail",
                rtype="thumbnail",
                role="object",
            )
            self._check_thumbnail_relations_and_schema(mapper)

    def test_targetentitymapper_relations_and_schema(self):
        with self.admin_access.cnx() as cnx:
            thumbnail = cnx.create_entity("Thumbnail", data=Binary(b"plip"))
            cnx.create_entity(
                "Photo", data=Binary(b"plop"), thumbnail=thumbnail
            )
            cnx.commit()
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.entity",
                cnx,
                entity=thumbnail,
                rtype="thumbnail",
                role="object",
            )
            self._check_thumbnail_relations_and_schema(mapper)

    def _check_thumbnail_relations_and_schema(self, mapper):
        relations = [
            (rtype, role) for rtype, role, _ in mapper.relations("view")
        ]
        self.assertEqual([("data", "subject")], relations)
        relations = [
            (rtype, role) for rtype, role, _ in mapper.relations("creation")
        ]
        self.assertEqual([("data", "subject")], relations)
        schema = mapper.jsl_document(VIEW_ROLE).get_schema()
        expected = {
            "$schema": "http://json-schema.org/draft-06/schema#",
            "title": "Thumbnail",
            "type": "object",
            "properties": {
                "data": {
                    "title": "data",
                    "type": "string",
                },
            },
            "additionalProperties": False,
        }
        self.assertEqual(schema, expected)

    def test_relation_targets(self):
        with self.admin_access.cnx() as cnx:
            dinosaurs = cnx.create_entity("Topic", name="Dinosaurs")
            monsters = cnx.create_entity("Topic", name="Monsters")
            programming = cnx.create_entity("Topic", name="Programming")
            book = cnx.create_entity(
                "Book", title="Creatures", topics=[dinosaurs, monsters]
            )

            mapper = cnx.vreg["mappers"].select(
                "jsonschema.relation",
                cnx,
                entity=book,
                rtype="topics",
                role="subject",
                target_types={"Topic"},
            )

            # Only related entities appear in view schema
            view_targets = [
                target.eid for target in mapper.relation_targets(VIEW_ROLE)
            ]
            self.assertIn(dinosaurs.eid, view_targets)
            self.assertIn(monsters.eid, view_targets)
            self.assertNotIn(programming.eid, view_targets)

            # All entities that could be related appear in edition schema
            targets = [
                target.eid for target in mapper.relation_targets(EDITION_ROLE)
            ]
            self.assertIn(dinosaurs.eid, targets)
            self.assertIn(monsters.eid, targets)
            self.assertIn(programming.eid, targets)

            # Only unrelated entities appear in creation schema
            targets = [
                target.eid for target in mapper.relation_targets(CREATION_ROLE)
            ]
            self.assertNotIn(dinosaurs.eid, targets)
            self.assertNotIn(monsters.eid, targets)
            self.assertIn(programming.eid, targets)

    def test_schema(self):
        expected = {
            "$schema": "http://json-schema.org/draft-06/schema#",
            "title": "Photo",
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "data": {
                    "title": "data",
                    "type": "string",
                },
                "exif_data": {
                    "$ref": "#/definitions/EXIF data",
                },
                "media_type": {
                    "title": "media_type",
                    "type": "string",
                },
                "thumbnail": {
                    "items": {
                        "$ref": "#/definitions/Thumbnail",
                    },
                    "title": "thumbnail",
                    "type": "array",
                },
            },
            "definitions": {
                "EXIF data": {
                    "additionalProperties": False,
                    "properties": {
                        "exposure_time": {
                            "title": "exposure_time",
                            "type": "number",
                        },
                        "flash": {
                            "title": "flash",
                            "type": "boolean",
                        },
                        "maker_note": {
                            "title": "maker_note",
                            "type": "string",
                        },
                    },
                    "title": "EXIF data",
                    "type": "object",
                },
                "Thumbnail": {
                    "additionalProperties": False,
                    "properties": {
                        "data": {
                            "title": "data",
                            "type": "string",
                        },
                    },
                    "title": "Thumbnail",
                    "type": "object",
                },
            },
        }
        with self.admin_access.cnx() as cnx:
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.entity", cnx, etype="Photo"
            )
            view_schema = mapper.jsl_document("view").get_schema()
            self.assertEqual(view_schema, expected)
            creation_schema = mapper.jsl_document("creation").get_schema()

            # Update expected schema to insert constraints.
            expected["required"] = ["data", "media_type"]
            expected["definitions"]["EXIF data"]["properties"]["flash"][
                "default"
            ] = False
            expected["definitions"]["EXIF data"]["required"] = ["flash"]
            expected["properties"]["media_type"]["default"] = "png"
            del expected["properties"]["media_type"]["type"]
            expected["properties"]["media_type"]["oneOf"] = [
                {"enum": ["jpeg"], "title": "jpeg", "type": "string"},
                {"enum": ["png"], "title": "png", "type": "string"},
            ]
            expected["properties"]["thumbnail"]["maxItems"] = 1
            expected["properties"]["thumbnail"]["minItems"] = 0
            expected["definitions"]["Thumbnail"]["required"] = ["data"]
            self.assertEqual(creation_schema, expected)

    def test_values_creation(self):
        with self.admin_access.cnx() as cnx:
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.entity", cnx, etype="Photo"
            )
            instance = {"data": "plip"}
            values = mapper.values(instance)
            self.assertEqual(list(values), ["data"])
            self.assertEqual(values["data"].getvalue(), b"plip")
            instance = {
                "data": "plip",
                "thumbnail": [
                    {
                        "data": "plop",
                    },
                ],
            }
            values = mapper.values(instance)
            self.assertCountEqual(list(values), ["data", "thumbnail"])
            self.assertEqual(values["data"].getvalue(), b"plip")
            (thumbnail,) = values["thumbnail"]
            self.assertEqual(thumbnail.cw_etype, "Thumbnail")
            # Custom "base64" mapper for Thumbnail's data attribute.
            self.assertEqual(
                thumbnail.data.getvalue(), base64.b64decode(b"plop")
            )

    def test_values_update(self):
        with self.admin_access.cnx() as cnx:
            entity = cnx.create_entity(
                "Photo", data=Binary(b"plop"), flash=True, exposure_time=1.23
            )
            cnx.create_entity(
                "Thumbnail", data=Binary(b"plip"), reverse_thumbnail=entity
            )
            cnx.commit()
            mapper = cnx.vreg["mappers"].select(
                "jsonschema.entity", cnx, entity=entity
            )
            instance = {"data": "plip", "media_type": "jpeg"}
            values = mapper.values(instance.copy())
            # All subject relations should appear in "values"
            subjrels = [
                rschema.type
                for rschema in list(entity.e_schema.subject_relations.values())
                if not rschema.meta
            ]
            self.assertCountEqual(list(values), subjrels)
            self.assertEqual(values["data"].getvalue(), b"plip")
            self.assertEqual(values["media_type"], "jpeg")
            # These are absent from instance, so should be reset to None.
            self.assertIsNone(values["exposure_time"])
            self.assertIsNone(values["maker_note"])
            self.assertEqual(values["thumbnail"], [])
            # "flash" has a default value.
            self.assertEqual(values["flash"], False)
            # inlined-related entities should have been dropped.
            self.assertFalse(cnx.find("Thumbnail"))


class PredicatesTC(CubicWebTC):

    def test_yams_match_base(self):
        predicate = yams_match(etype="etype", rtype="rtype", role="role")

        self.assertEqual(
            predicate(
                None,
                None,
                etype="etype",
                rtype="rtype",
                role="role",
                target_types={"CWGroup"},
            ),
            3,
        )
        self.assertEqual(
            predicate(
                None,
                None,
                etype="notetype",
                rtype="rtype",
                role="role",
                target_types={"CWGroup"},
            ),
            0,
        )
        self.assertEqual(
            predicate(
                None,
                None,
                etype="etype",
                rtype="notrtype",
                role="role",
                target_types={"CWGroup"},
            ),
            0,
        )
        self.assertEqual(
            predicate(
                None,
                None,
                etype="etype",
                rtype="rtype",
                role="notrole",
                target_types={"CWGroup"},
            ),
            0,
        )

    def test_yams_match_entity(self):
        predicate = yams_match(etype="CWUser")
        with self.admin_access.client_cnx() as cnx:
            bob = cnx.create_entity("CWUser", upassword="123", login="bob")
            self.assertEqual(
                predicate(
                    None,
                    None,
                    entity=bob,
                    rtype="rtype",
                    role="role",
                    target_types=set([]),
                ),
                1,
            )
        predicate = yams_match(
            etype="CWUser",
            rtype="upassword",
            role="subject",
            target_types="Bytes",
        )
        with self.admin_access.client_cnx() as cnx:
            bob = cnx.create_entity("CWUser", upassword="123", login="bob")
            self.assertEqual(
                predicate(
                    None,
                    None,
                    entity=bob,
                    rtype="upassword",
                    role="subject",
                    target_types={"Bytes"},
                ),
                4,
            )

    def test_yams_match_target_types(self):
        predicate = yams_match(target_types="String")

        self.assertEqual(
            predicate(
                None,
                None,
                etype="etype",
                rtype="rtype",
                role="role",
                target_types={"CWGroup"},
            ),
            0,
        )
        self.assertEqual(
            predicate(
                None,
                None,
                etype="etype",
                rtype="rtype",
                role="role",
                target_types={"String"},
            ),
            1,
        )
        predicate = yams_match(target_types={"CWUser", "CWGroup"})
        self.assertEqual(
            predicate(
                None,
                None,
                etype="etype",
                rtype="rtype",
                role="role",
                target_types={"CWUser", "CWGroup"},
            ),
            1,
        )
        self.assertEqual(
            predicate(
                None,
                None,
                etype="etype",
                rtype="rtype",
                role="role",
                target_types={"CWUser"},
            ),
            1,
        )

    def test_yams_match_all(self):
        predicate = yams_match(
            etype="Photo", rtype="data", role="subject", target_types="Bytes"
        )

        self.assertEqual(
            predicate(
                None,
                None,
                etype="etype",
                rtype="rtype",
                role="role",
                target_types={"CWGroup"},
            ),
            0,
        )
        self.assertEqual(
            predicate(
                None,
                None,
                etype="Thumbnail",
                rtype="data",
                role="subject",
                target_types={"Bytes"},
            ),
            0,
        )
        self.assertEqual(
            predicate(
                None,
                None,
                etype="Photo",
                rtype="data",
                role="subject",
                target_types={"Bytes"},
            ),
            4,
        )


if __name__ == "__main__":
    unittest.main()
